-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THMoistureSystem = {}
local THMoistureSystem_mt = THUtils.createClass(THMoistureSystem, THAbstractSystem)
local debugFlagId = THDebugUtil.createDebugFlag("THMoistureSystem")
THMoistureSystem.debugFlagId = debugFlagId
THMoistureSystem.DATA_KEY = "moistureSystem"
THMoistureSystem.FACTOR_SEND_BITS = 16
THMoistureSystem.PERIOD_SEND_BITS = 6
THMoistureSystem.DEFAULT_DRYING_FACTORS = {
    NIGHT = 0.5,
    CLOUD = 0.5,
    RAIN  = 0,
    STORM = 0,
    SNOW = 0.25,
    WIND = 1
}
function THMoistureSystem.initXMLSchema(xmlSchema, xmlPath)
    THMoistureSystem:superClass().initXMLSchema(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath .. ".environment", "#minHumidity", "Absolute minimum allowed humidity percentage", 30)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath .. ".environment", "#dryingFactor", "How fast environment dries after rain", 1)
    local function registerFruitTypeDryingStates(pBasePath)
        local dryingStatePath = pBasePath .. ".dryingStates.state(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, dryingStatePath, "#min", "Minimum fruitType growth state")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, dryingStatePath, "#max", "Maximum fruitType growth state")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#pctStart", "FruitType starting moisture percentage", 100)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#pctEnd", "FruitType ending moisture percentage", 0)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#hoursToDry", "Hours for crop to dry under optimal conditions")
        local resetPeriodPath = dryingStatePath .. ".resetPeriods.period(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, resetPeriodPath, "#id", "Game period id/index")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, resetPeriodPath, "reset", "Drying state reset flag", true)
        local dryingFactorsPath = dryingStatePath .. ".baseFactors"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#night", "Amount that night affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.NIGHT)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#cloud", "Amount that clouds affect dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.CLOUD)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#rain", "Amount that rain affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.RAIN)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#storm", "Amount that hail/storms affect dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.STORM)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#snow", "Amount that snow affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.SNOW)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#wind", "Amount that wind affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.WIND)
        local humidityFactorPath = dryingStatePath .. ".humidityFactors.humidity(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, humidityFactorPath, "#value", "Humidity percentage")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, humidityFactorPath, "#factor", "Drying factor at this humidity percentage")
        local soilTypeFactorPath = dryingStatePath .. ".soilTypeFactors.soilType(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, soilTypeFactorPath, "#index", "SoilType index")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, soilTypeFactorPath, "factor", "SoilType factor")
    end
    local fruitTypePath = xmlPath .. ".fruitTypes.fruitType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fruitTypePath, "#name", "FruitType name")
    registerFruitTypeDryingStates(fruitTypePath)
    local fillTypePath = xmlPath .. ".fillTypes.fillType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fillTypePath, "#name", "FillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypePath, "#min", "Mininum moisture percentage")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypePath, "#max", "Maximum moisture percentage")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypePath, "#default", "Default moisture percentage")
    local fillTypeFactorPath = fillTypePath .. ".factors.factor(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#moisture", "FillType moisture percentage")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#price", "Price factor at specified moisture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#yield", "Yield factor at specified moisture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#lossPerMonth", "Storage loss per in-game month at specified moisture")
end
function THMoistureSystem.initXMLSchemaSavegame(xmlSchema, xmlPath)
    THMoistureSystem:superClass().initXMLSchemaSavegame(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#cachedDryingTime", "Cached drying time", 0)
    local dryingConditionsPath = xmlPath .. ".dryingConditions"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, dryingConditionsPath, "#isRaining", "Last isRaining value", false)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, dryingConditionsPath, "#isDayTime", "Last isSunOn value", false)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingConditionsPath, "#temp", "Last temperature value", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingConditionsPath, "#dewPoint", "Last dew point value", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, dryingConditionsPath, "#periodLength", "Current daysPerPeriod setting", 1)
    local currentRainPath = dryingConditionsPath .. ".currentRain"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#startDay", "Game day when rain event started", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#startTime", "Game time of day when rain event started", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#endDay", "Game day when rain event ended", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#endTime", "Game time of day when rain event ended", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#duration", "Total game time that rain event lasted", 0)
    local lastRainPath = dryingConditionsPath .. ".lastRain"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#startDay", "Game day when rain event started", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#startTime", "Game time of day when rain event started", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#endDay", "Game day when rain event ended", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#endTime", "Game time of day when rain event ended", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#duration", "Total game time that rain event lasted", 0)
    local fruitTypePath = xmlPath .. ".fruitTypes.fruitType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fruitTypePath, "#name", "FruitType name")
    local dryingStatePath = fruitTypePath .. ".dryingStates.dryingState(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, dryingStatePath, "#index", "Drying state index")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#averageTemp", "Average temperature", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#averageWind", "Average wind speed", 0)
    local dryingTimesPath = dryingStatePath .. ".dryingTimes"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#total", "Total drying hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#night", "Total night hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#cloud", "Total cloud hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#rain", "Total rain hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#storm", "Total storm hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#snow", "Total snow hours", 0)
    local farmlandFactorPath = dryingStatePath .. ".farmlandFactors.farmland(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, farmlandFactorPath, "#id", "Farmland id")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, farmlandFactorPath, "#factor", "Farmland drying factor")
end
function THMoistureSystem.new(owner, dataKey, modName, customMt)
    customMt = customMt or THMoistureSystem_mt
    local self = THAbstractSystem.new(owner, dataKey, modName, customMt)
    self.thSettingsManager = owner.thSettingsManager
    self.thSpecManager = owner.thSpecManager
    self.thFruitTypeManager = owner.thFruitTypeManager
    self.thFillTypeManager = owner.thFillTypeManager
    self.thPFManager = owner.thPFManager
    self.dryingConditions = {
        isDayTime = false,
        isRaining = false,
        humidity = 0,
        dewPoint = 0,
        temperature = 0,
        absMinHumidity = 0.3,
        absMaxHumidity = 1,
        maxDryingTime = 720, -- 12 in-game hours before/after a rain
        dryingFactor = 1,
        periodLength = 1,
        dryingPeriods = {},
        isInitialized = false
    }
    self.dryingConditions.lastRainData = {
        startTime = 0,
        endTime = 0,
        startDay = 1,
        endDay = 1,
        duration = 0
    }
    self.dryingConditions.currentRainData = {
        startTime = 0,
        endTime = 0,
        startDay = 1,
        endDay = 1,
        duration = 0
    }
    self.fruitTypes = {}
    self.dryingStateCachedTime = 0
    self.isDryingStateUpdateSafe = false
    self.areDryingStatesInitialized = false
    self.resetPeriodIndexToId = {}
    for periodId, periodIndex in pairs(Environment.PERIOD) do
        self.resetPeriodIndexToId[periodIndex] = periodId
    end
    self.fillTypes = {}
    self.moneyChangeTypes = {
        harvestMoisture = MoneyType.register("harvestIncome", "thFinance_harvestMoisture", modName),
        dryingCosts = MoneyType.register("harvestIncome", "thFinance_dryingCosts", modName)
    }
    self.dirtyFlag = self:getNextDirtyFlag()
    self.messageCenter:subscribe(MessageType.MINUTE_CHANGED, self.onMinuteChanged, self)
    self.messageCenter:subscribe(MessageType.HOUR_CHANGED, self.onHourChanged, self)
    self.messageCenter:subscribe(MessageType.DAY_CHANGED, self.onDayChanged, self)
    self.messageCenter:subscribe(MessageType.PERIOD_CHANGED, self.onPeriodChanged, self)
    self.messageCenter:subscribe(MessageType.WEATHER_CHANGED, self.onWeatherChanged, self)
    return self
end
function THMoistureSystem.newModule(owner, modData)
    if _G.g_thMoistureSystem ~= nil then
        THUtils.warningMsg("Module (%s) already loaded", THMoistureSystem.DATA_KEY)
    else
        local self = THMoistureSystem.new(owner, THMoistureSystem.DATA_KEY, modData.name)
        if self:addSchemaFunctions() then
            _G.g_thMoistureSystem = self
            local modName = self.modName
            local modPath = self.modPath
            source(modPath .. "scripts/gui/THInGameMoistureWeatherFrame.lua")
            source(modPath .. "scripts/gui/THFillLevelMoistureDisplay.lua")
            source(modPath .. "scripts/gui/THPlayerMoistureHud.lua")
            source(modPath .. "scripts/objects/THMoistureStorage.lua")
            source(modPath .. "scripts/objects/THMoistureUnloadingStation.lua")
            source(modPath .. "scripts/objects/THMoistureLoadingStation.lua")
            source(modPath .. "scripts/objects/THStorageDryer.lua")
            local thSpecManager = self.thSpecManager
            thSpecManager:addSpecialization("vehicle", "moistureFillUnit", "THMoistureFillUnit", modName, "scripts/vehicles/specializations/THMoistureFillUnit.lua")
            thSpecManager:addSpecialization("vehicle", "moistureDischargeable", "THMoistureDischargeable", modName, "scripts/vehicles/specializations/THMoistureDischargeable.lua")
            thSpecManager:addSpecialization("vehicle", "moistureCombine", "THMoistureCombine", modName, "scripts/vehicles/specializations/THMoistureCombine.lua")
            thSpecManager:addSpecialization("placeable", "moistureStorage", "THPlaceableMoistureStorage", modName, "scripts/placeables/specializations/THPlaceableMoistureStorage.lua")
            return self
        end
    end
end
function THMoistureSystem:initializeDataTables()
    if self:initFillTypeData() and self:initFruitTypeData() then
        return true
    end
    return false
end
function THMoistureSystem:initFruitTypeData()
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypesArray, numFruitTypes = thFruitTypeManager:getFruitTypeList(true)
    if fruitTypesArray ~= nil and numFruitTypes > 0 then
        for fruitTypeIndex = 1, numFruitTypes do
            local fruitTypeInfo = fruitTypesArray[fruitTypeIndex]
            local fruitTypeData = thFruitTypeManager:getFruitTypeData(fruitTypeIndex)
            local fruitTypeMoistureData = {
                dryingStates = {},
                dryingStateMapping = {},
                isEnabled = true,
                needSync = true,
            }
            fruitTypeMoistureData.fruitType = {
                name = fruitTypeInfo.name,
                index = fruitTypeInfo.index,
                info = fruitTypeInfo,
                data = fruitTypeData
            }
            if fruitTypeData == nil then
                fruitTypeMoistureData.isEnabled = false
            end
            self.fruitTypes[fruitTypeIndex] = fruitTypeMoistureData
        end
    end
    return true
end
function THMoistureSystem:initFillTypeData()
    local thFillTypeManager = self.thFillTypeManager
    local fillTypesArray, numFillTypes = thFillTypeManager:getFillTypeList(true)
    if fillTypesArray ~= nil and numFillTypes > 0 then
        for fillTypeIndex = 1, numFillTypes do
            local fillTypeInfo = fillTypesArray[fillTypeIndex]
            local fillTypeData = thFillTypeManager:getFillTypeData(fillTypeIndex)
            local fillTypeMoistureData = {
                minPercent = 0,
                maxPercent = 1,
                defPercent = 0.5,
                isEnabled = true,
                hasMoisture = false,
                needSync = true
            }
            fillTypeMoistureData.fillType = {
                name = fillTypeInfo.name,
                title = fillTypeInfo.title,
                index = fillTypeInfo.index,
                info = fillTypeInfo,
                data = fillTypeData
            }
            fillTypeMoistureData.factors = {
                curve = {},
                minPrice = 1,
                maxPrice = 1,
                minYield = 1,
                maxYield = 1,
                minDecay = 0,
                maxDecay = 0
            }
            if fillTypeData == nil then
                fillTypeMoistureData.isEnabled = false
            end
            self.fillTypes[fillTypeIndex] = fillTypeMoistureData
        end
    end
    return true
end
function THMoistureSystem:finalizeDataTables()
    if self:finalizeFruitTypeData() and self:finalizeFillTypeData() then
        if THDebugUtil.getIsEnabled(debugFlagId) then
            THDebugUtil.printMsg(debugFlagId, nil, "Moisture System Environment")
            THDebugUtil.printTable(debugFlagId, nil, self, 0)
            THDebugUtil.printMsg(debugFlagId, nil, "Drying Conditions:")
            THDebugUtil.printTable(debugFlagId, nil, self.dryingConditions, 5)
            for _, fruitTypeMoistureData in ipairs(self.fruitTypes) do
                THDebugUtil.printMsg(debugFlagId, nil, "Moisture values for fruitType: %s", fruitTypeMoistureData.fruitType.name)
                THDebugUtil.printTable(debugFlagId, nil, fruitTypeMoistureData, 0)
                THDebugUtil.printMsg(debugFlagId, nil, "Drying States:")
                for _, dryingStateData in ipairs(fruitTypeMoistureData.dryingStates) do
                    THDebugUtil.printTable(debugFlagId, nil, dryingStateData, 0)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Drying Times:")
                    THDebugUtil.printTable(debugFlagId, nil, dryingStateData.dryingTimes, 5)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Drying Factors:")
                    THDebugUtil.printTable(debugFlagId, nil, dryingStateData.dryingFactors, 5)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Reset Periods:")
                    THDebugUtil.printTable(debugFlagId, nil, dryingStateData.resetPeriods, 5)
                end
            end
            for _, fillTypeMoistureData in ipairs(self.fillTypes) do
                THDebugUtil.printMsg(debugFlagId, nil, "Moisture values for fillType: %s", fillTypeMoistureData.fillType.name)
                THDebugUtil.printTable(debugFlagId, nil, fillTypeMoistureData, 0)
                THDebugUtil.printMsg(debugFlagId, nil, "- Moisture factors:")
                THDebugUtil.printTable(debugFlagId, nil, fillTypeMoistureData.factors, 5)
            end
        end
        return true
    end
    return false
end
function THMoistureSystem:finalizeFruitTypeData()
    local thFruitTypeManager = self.thFruitTypeManager
    local growthSystem = self.growthSystem
    local maxNumPeriods = Environment.PERIODS_IN_YEAR
    local autoConfiguredCrops = {}
    for fruitTypeIndex, fruitTypeMoistureData in ipairs(self.fruitTypes) do
        local fruitTypeData = fruitTypeMoistureData.fruitType.data
        local fillTypeInfo = thFruitTypeManager:getFruitTypeFillType(fruitTypeIndex)
        local fillTypeMoistureData = self:getFillTypeMoistureData(fillTypeInfo, true) -- force
        THUtils.clearTable(fruitTypeMoistureData.dryingStateMapping) -- safety
        if fruitTypeData == nil or fillTypeInfo == nil or fillTypeMoistureData == nil then
            THUtils.clearTable(fruitTypeMoistureData.dryingStates)
            fruitTypeMoistureData.isEnabled = false
        else
            local minHarvestState = fruitTypeData.minHarvestState or -1
            local maxHarvestState = fruitTypeData.maxHarvestState or -1
            local hasDryingStates = #fruitTypeMoistureData.dryingStates > 0
            if minHarvestState < 0 or maxHarvestState < 0 then
                THUtils.clearTable(fruitTypeMoistureData.dryingStates)
                hasDryingStates = false
            elseif not hasDryingStates then
                local dryingStateData = self:createFruitTypeDryingState(fruitTypeIndex, minHarvestState, maxHarvestState, 0.3, 0.1)
                if dryingStateData ~= nil then
                    dryingStateData.index = #fruitTypeMoistureData.dryingStates + 1
                    fruitTypeMoistureData.dryingStates[dryingStateData.index] = dryingStateData
                    autoConfiguredCrops[fruitTypeMoistureData] = true
                    hasDryingStates = true
                end
            end
            if hasDryingStates then
                local minHarvestMoisture = math.huge
                local maxHarvestMoisture = -math.huge
                for dryingStateIndex, dryingStateData in ipairs(fruitTypeMoistureData.dryingStates) do
                    local minGrowthState = dryingStateData.minGrowthState
                    local maxGrowthState = dryingStateData.maxGrowthState
                    local resetPeriods = dryingStateData.resetPeriods
                    dryingStateData.index = dryingStateIndex
                    if minGrowthState >= minHarvestState or maxGrowthState <= maxHarvestState then
                        minHarvestMoisture = math.min(minHarvestMoisture, dryingStateData.startPercent, dryingStateData.endPercent)
                        maxHarvestMoisture = math.max(maxHarvestMoisture, dryingStateData.startPercent, dryingStateData.endPercent)
                    end
                    for growthState = minGrowthState, maxGrowthState do
                        fruitTypeMoistureData.dryingStateMapping[growthState] = dryingStateIndex
                    end
                    if maxNumPeriods == 0 then
                        resetPeriods = THUtils.clearTable(dryingStateData.resetPeriods)
                        dryingStateData.hasResetPeriods = false
                    elseif dryingStateData.hasResetPeriods and next(resetPeriods) == nil then
                        local growthData = growthSystem.seasonalFruitData[fruitTypeIndex]
                        dryingStateData.hasResetPeriods = false -- reset
                        if growthData ~= nil then
                            for period = 1, maxNumPeriods do
                                local periodData = growthData[period]
                                if periodData ~= nil and periodData.harvestPossible then
                                    local lastPeriodData = nil -- placeholder
                                    dryingStateData.hasResetPeriods = true
                                    if period == 1 then
                                        lastPeriodData = growthData[maxNumPeriods]
                                    else
                                        lastPeriodData = growthData[period - 1]
                                    end
                                    if lastPeriodData == nil or not lastPeriodData.harvestPossible then
                                        resetPeriods[period] = true
                                    end
                                end
                            end
                            if dryingStateData.hasResetPeriods then
                                if next(resetPeriods) == nil then
                                    for period = 1, maxNumPeriods do
                                        resetPeriods[period] = true
                                    end
                                end
                                autoConfiguredCrops[fruitTypeMoistureData] = true
                            end
                        end
                    end
                end
                if fillTypeMoistureData.isEnabled
                    and not fillTypeMoistureData.hasMoisture
                    and minHarvestMoisture < math.huge and maxHarvestMoisture > -math.huge
                then
                    THUtils.clearTable(fillTypeMoistureData.factors.curve)
                    if self:setFillTypeMoistureFactors(fillTypeInfo, minHarvestMoisture, 1, 1, 0)
                        and self:setFillTypeMoistureFactors(fillTypeInfo, maxHarvestMoisture, 0.8, 0.8, 0.05)
                    then
                        fillTypeMoistureData.minPercent = minHarvestMoisture
                        fillTypeMoistureData.maxPercent = maxHarvestMoisture
                        fillTypeMoistureData.defPercent = (minHarvestMoisture + maxHarvestMoisture) / 2
                        fillTypeMoistureData.hasMoisture = true
                        autoConfiguredCrops[fruitTypeMoistureData] = true
                    end
                end
            end
        end
    end
    for fruitTypeMoistureData in pairs(autoConfiguredCrops) do
        THUtils.displayMsg("Moisture values have been auto-generated for fruitType: %s", fruitTypeMoistureData.fruitType.name)
    end
    return true
end
function THMoistureSystem:finalizeFillTypeData()
    for _, fillTypeMoistureData in ipairs(self.fillTypes) do
        local moistureFactors = fillTypeMoistureData.factors
        local numCurveEntries = #moistureFactors.curve
        if numCurveEntries > 0 then
            local minPrice = math.huge
            local minYield = math.huge
            local minDecay = math.huge
            local maxPrice = -math.huge
            local maxYield = -math.huge
            local maxDecay = -math.huge
            for curveIndex = 1, numCurveEntries do
                local curveEntry = moistureFactors.curve[curveIndex]
                local priceFactor = THUtils.getNoNil(curveEntry.priceFactor, 1)
                local yieldFactor = THUtils.getNoNil(curveEntry.yieldFactor, 1)
                local decayFactor = THUtils.getNoNil(curveEntry.decayFactor, 0)
                priceFactor = math.max(0, priceFactor)
                yieldFactor = math.max(0, yieldFactor)
                decayFactor = math.max(0, decayFactor)
                curveEntry.priceFactor = priceFactor
                curveEntry.yieldFactor = yieldFactor
                curveEntry.decayFactor = decayFactor
                minPrice = math.min(minPrice, priceFactor)
                minYield = math.min(minYield, yieldFactor)
                minDecay = math.min(minDecay, decayFactor)
                maxPrice = math.max(maxPrice, priceFactor)
                maxYield = math.max(maxYield, yieldFactor)
                maxDecay = math.max(maxDecay, decayFactor)
            end
            moistureFactors.minPrice = math.min(minPrice, maxPrice)
            moistureFactors.minYield = math.min(minYield, maxYield)
            moistureFactors.minDecay = math.min(minDecay, maxDecay)
            moistureFactors.maxPrice = math.max(minPrice, maxPrice)
            moistureFactors.maxYield = math.max(minYield, maxYield)
            moistureFactors.maxDecay = math.max(minDecay, maxDecay)
        end
    end
    return true
end
function THMoistureSystem:loadFromConfigXMLFile(xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local thFillTypeManager = self.thFillTypeManager
    local loadSuccess = THMoistureSystem:superClass().loadFromConfigXMLFile(self, xmlFile, xmlKey)
    if loadSuccess then
        local dryingConditions = self.dryingConditions
        local environmentKey = xmlKey .. ".environment"
        local absMinHumidity = THUtils.getXMLValue(xmlFile, environmentKey, "#minHumidity", 30)
        local dryingFactor = THUtils.getXMLValue(xmlFile, environmentKey, "#dryingFactor", 1)
        if absMinHumidity < 0 or absMinHumidity > 100 then
            THUtils.xmlLoadMsg(environmentKey, "minHumidity", absMinHumidity, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
        else
            dryingConditions.absMinHumidity = MathUtil.clamp(absMinHumidity / 100, 0, 1)
        end
        if dryingFactor <= 0 then
            THUtils.xmlLoadMsg(environmentKey, "dryingFactor", dryingFactor, true, THXMLMessage.VALUE_GREATER_ZERO)
        else
            dryingConditions.dryingFactor = dryingFactor
        end
    end
    if loadSuccess then
        xmlFile:iterate(xmlKey .. ".fillTypes.fillType", function(_, pFillTypeKey)
            local fillTypeName = THUtils.getXMLValue(xmlFile, pFillTypeKey, "#name")
            local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
            if fillTypeInfo == nil or fillTypeInfo.index == FillType.UNKNOWN then
                THUtils.xmlLoadMsg(pFillTypeKey, "name", fillTypeName, fillTypeName == nil, THXMLMessage.INVALID_FILL_TYPE)
            else
                if not self:loadFillTypeMoisture(fillTypeInfo.index, xmlFile, pFillTypeKey) then
                    loadSuccess = false
                    return false
                end
            end
        end)
    end
    if loadSuccess then
        xmlFile:iterate(xmlKey .. ".fruitTypes.fruitType", function(_, pFruitTypeKey)
            local fruitTypeName = THUtils.getXMLValue(xmlFile, pFruitTypeKey, "#name")
            local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeName)
            if fruitTypeInfo == nil or fruitTypeInfo.index == FruitType.UNKNOWN then
                THUtils.xmlLoadMsg(pFruitTypeKey, "name", fruitTypeName, fruitTypeName == nil, THXMLMessage.INVALID_FRUIT_TYPE)
            else
                if not self:loadFruitTypeMoisture(fruitTypeInfo.index, xmlFile, pFruitTypeKey) then
                    loadSuccess = false
                    return false
                end
            end
        end)
    end
    return loadSuccess
end
function THMoistureSystem:loadFillTypeMoisture(fillType, xmlFile, xmlKey)
    local thFillTypeManager = self.thFillTypeManager
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType, true) -- force
    local minPercent = THUtils.getXMLValue(xmlFile, xmlKey, "#min")
    local maxPercent = THUtils.getXMLValue(xmlFile, xmlKey, "#max")
    local defPercent = THUtils.getXMLValue(xmlFile, xmlKey, "#default")
    local loadSuccess = true
    if fillTypeMoistureData == nil then
        local fillTypeInfo = thFillTypeManager:getFillType(fillType)
        if fillTypeInfo ~= nil then
            THUtils.warningMsg("Could not find moisture data for fillType (%s)", fillTypeInfo.name)
        end
    elseif fillTypeMoistureData.isEnabled then
        local isMoistureValid = false
        if minPercent ~= nil or maxPercent ~= nil then
            minPercent = THUtils.getNoNil(minPercent, 0)
            maxPercent = THUtils.getNoNil(maxPercent, 100)
            isMoistureValid = true -- activate moisture for this fillType
            if minPercent < 0 or minPercent > 100 then
                THUtils.xmlLoadMsg(xmlKey, "min", minPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                isMoistureValid = false
            elseif maxPercent < 0 or maxPercent > 100 then
                THUtils.xmlLoadMsg(xmlKey, "max", maxPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                isMoistureValid = false
            elseif minPercent > maxPercent then
                THUtils.xmlLoadMsg(xmlKey, "max", maxPercent, true, "is less than min (%s)", minPercent)
                isMoistureValid = false
            else
                fillTypeMoistureData.minPercent = MathUtil.clamp(minPercent / 100, 0, 1)
                fillTypeMoistureData.maxPercent = MathUtil.clamp(maxPercent / 100, 0, 1)
            end
        end
        local xmlMinPercent = fillTypeMoistureData.minPercent * 100
        local xmlMaxPercent = fillTypeMoistureData.maxPercent * 100
        if isMoistureValid then
            if defPercent ~= nil then
                if defPercent < 0 or defPercent > 100 then
                    THUtils.xmlLoadMsg(xmlKey, "default", defPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                    isMoistureValid = false
                elseif defPercent < xmlMinPercent then
                    THUtils.xmlLoadMsg(xmlKey, "default", defPercent, true, "is less than minimum moisture (%s)", xmlMinPercent)
                    isMoistureValid = false
                elseif defPercent > xmlMaxPercent then
                    THUtils.xmlLoadMsg(xmlKey, "default", defPercent, true, "is greater than maximum moisture (%s)", xmlMaxPercent)
                    isMoistureValid = false
                else
                    fillTypeMoistureData.defPercent = MathUtil.clamp(defPercent / 100, 0, 1)
                end
            end
            if not isMoistureValid or defPercent == nil then
                fillTypeMoistureData.defPercent = (fillTypeMoistureData.minPercent + fillTypeMoistureData.maxPercent) / 2
            end
        end
        if isMoistureValid then
            local moistureFactors = fillTypeMoistureData.factors
            local isCurveReset = false
            xmlFile:iterate(xmlKey .. ".factors.factor", function(_, pFactorKey)
                local moisture = THUtils.getXMLValue(xmlFile, pFactorKey, "#moisture")
                local priceFactor = THUtils.getXMLValue(xmlFile, pFactorKey, "#price", 1)
                local yieldFactor = THUtils.getXMLValue(xmlFile, pFactorKey, "#yield", 1)
                local decayFactor = THUtils.getXMLValue(xmlFile, pFactorKey, "#decayPerMonth", 0)
                if moisture == nil then
                    THUtils.xmlLoadMsg(pFactorKey, "moisture", moisture, true, THXMLMessage.INVALID_VALUE)
                elseif moisture < 0 or moisture > 100 then
                    THUtils.xmlLoadMsg(pFactorKey, "moisture", moisture, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                elseif priceFactor < 0 then
                    THUtils.xmlLoadMsg(pFactorKey, "price", priceFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                elseif yieldFactor < 0 then
                    THUtils.xmlLoadMsg(pFactorKey, "yield", yieldFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                elseif decayFactor < 0 or decayFactor > 1 then
                    THUtils.xmlLoadMsg(pFactorKey, "decayPerMonth", decayFactor, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 1)
                else
                    local realMoisture = MathUtil.clamp(moisture / 100, 0, 1)
                    if moisture < xmlMinPercent then
                        THUtils.xmlLoadMsg(pFactorKey, "moisture", moisture, true, "is less than minimum moisture (%s)", xmlMinPercent)
                    elseif moisture > xmlMaxPercent then
                        THUtils.xmlLoadMsg(pFactorKey, "moisture", moisture, true, "is greater than maximum moisture (%s)", xmlMaxPercent)
                    else
                        if not isCurveReset then
                            THUtils.clearTable(moistureFactors.curve)
                            isCurveReset = true
                        end
                        if not self:setFillTypeMoistureFactors(fillTypeMoistureData.fillType.index, realMoisture, priceFactor, yieldFactor, decayFactor) then
                            THUtils.stackMsg("Could not set moisture factors for fillType: %s", fillTypeMoistureData.fillType.name)
                            isMoistureValid = false
                            return false
                        end
                    end
                end
            end)
        end
        fillTypeMoistureData.hasMoisture = isMoistureValid == true
    end
    return loadSuccess
end
function THMoistureSystem:loadFruitTypeMoisture(fruitType, xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local pfManager = self.thPFManager
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    local loadSuccess = true
    if fruitTypeMoistureData ~= nil then
        local fruitTypeIndex = fruitTypeMoistureData.fruitType.index
        if xmlFile:hasProperty(xmlKey .. ".dryingStates") then
            THUtils.clearTable(fruitTypeMoistureData.dryingStates)
            xmlFile:iterate(xmlKey .. ".dryingStates.state", function(_, pDryingStateKey)
                local minStateId = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#min")
                local maxStateId = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#max")
                local pctStart = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#pctStart")
                local pctEnd = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#pctEnd")
                local hoursToDry = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#hoursToDry")
                minStateId = THUtils.getNoNil(minStateId, maxStateId, true)
                maxStateId = THUtils.getNoNil(maxStateId, minStateId, true)
                local minState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, minStateId, true)
                local maxState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, maxStateId, true)
                local isDryingStateValid = true
                if minState == nil or minState < 0 then
                    THUtils.xmlLoadMsg(pDryingStateKey, "min", minStateId, true, THXMLMessage.INVALID_VALUE)
                    isDryingStateValid = false
                elseif maxState == nil or maxState < 0 then
                    THUtils.xmlLoadMsg(pDryingStateKey, "max", maxStateId, true, THXMLMessage.INVALID_VALUE)
                    isDryingStateValid = false
                end
                if isDryingStateValid then
                    if pctStart == nil then
                        THUtils.xmlLoadMsg(pDryingStateKey, "pctStart", pctStart, true, THXMLMessage.INVALID_VALUE)
                        isDryingStateValid = false
                    elseif pctStart < 0 or pctStart > 100 then
                        THUtils.xmlLoadMsg(pDryingStateKey, "pctStart", pctStart, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                        isDryingStateValid = false
                    else
                        pctStart = pctStart / 100
                    end
                end
                if isDryingStateValid then
                    if pctEnd == nil then
                        THUtils.xmlLoadMsg(pDryingStateKey, "pctEnd", pctEnd, true, THXMLMessage.INVALID_VALUE)
                        isDryingStateValid = false
                    elseif pctEnd < 0 or pctEnd > 100 then
                        THUtils.xmlLoadMsg(pDryingStateKey, "pctEnd", pctEnd, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                        isDryingStateValid = false
                    else
                        pctEnd = pctEnd / 100
                    end
                end
                if isDryingStateValid then
                    local dryingStateData = self:createFruitTypeDryingState(fruitTypeIndex, minState, maxState, pctStart, pctEnd)
                    if dryingStateData == nil then
                        loadSuccess = false
                        return false
                    end
                    if hoursToDry ~= nil then
                        if hoursToDry < 0 then
                            THUtils.xmlLoadMsg(pDryingStateKey, "hoursToDry", hoursToDry, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                            isDryingStateValid = false
                        else
                            dryingStateData.timeToDry = hoursToDry * 60 -- convert to minutes to dry
                        end
                    end
                    local resetPeriods = dryingStateData.resetPeriods
                    local resetPeriodsKey = pDryingStateKey .. ".resetPeriods"
                    if isDryingStateValid and xmlFile:hasProperty(resetPeriodsKey) then
                        resetPeriods = THUtils.clearTable(dryingStateData.resetPeriods)
                        xmlFile:iterate(resetPeriodsKey .. ".period", function(_, pResetPeriodKey)
                            local resetPeriodId = THUtils.getXMLValue(xmlFile, pResetPeriodKey, "#id")
                            local isResetPeriod = THUtils.getXMLValue(xmlFile, pResetPeriodKey, "#reset", true)
                            if resetPeriodId == nil then
                                THUtils.xmlLoadMsg(pResetPeriodKey, "id", resetPeriodId, true, THXMLMessage.INVALID_VALUE)
                            else
                                local period = tonumber(resetPeriodId)
                                if period == nil or self.resetPeriodIndexToId[period] == nil then
                                    period = Environment.PERIOD[resetPeriodId:upper()]
                                end
                                if period == nil then
                                    THUtils.xmlLoadMsg(pResetPeriodKey, "id", resetPeriodId, true, THXMLMessage.INVALID_VALUE)
                                elseif isResetPeriod then
                                    resetPeriods[period] = true
                                else
                                    resetPeriods[period] = nil
                                end
                            end
                        end)
                        if next(resetPeriods) ~= nil then
                            dryingStateData.hasResetPeriods = true
                        end
                    end
                    local dryingFactors = dryingStateData.dryingFactors
                    if isDryingStateValid then
                        local dryingFactorsKey = pDryingStateKey .. ".baseFactors"
                        dryingFactors.night = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#night", THMoistureSystem.DEFAULT_DRYING_FACTORS.NIGHT)
                        dryingFactors.cloud = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#cloud", THMoistureSystem.DEFAULT_DRYING_FACTORS.CLOUD)
                        dryingFactors.rain = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#rain", THMoistureSystem.DEFAULT_DRYING_FACTORS.RAIN)
                        dryingFactors.storm = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#storm", THMoistureSystem.DEFAULT_DRYING_FACTORS.STORM)
                        dryingFactors.snow = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#snow", THMoistureSystem.DEFAULT_DRYING_FACTORS.SNOW)
                        dryingFactors.wind = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#wind", THMoistureSystem.DEFAULT_DRYING_FACTORS.WIND)
                    end
                    if isDryingStateValid then
                        local humidityFactorsKey = pDryingStateKey .. ".humidityFactors"
                        if xmlFile:hasProperty(humidityFactorsKey) then
                            local humidityFactors = THUtils.clearTable(dryingFactors.humidity)
                            xmlFile:iterate(humidityFactorsKey .. ".humidity", function(_, pHumidityFactorKey)
                                local humidityValue = THUtils.getXMLValue(xmlFile, pHumidityFactorKey, "#value")
                                local humidityFactor = THUtils.getXMLValue(xmlFile, pHumidityFactorKey, "#factor", 1)
                                if humidityValue == nil or humidityValue < 0 or humidityValue > 100 then
                                    THUtils.xmlLoadMsg(pHumidityFactorKey, "value", humidityValue, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                                else
                                    local humidityFactorData = {
                                        value = humidityValue / 100,
                                        factor = humidityFactor
                                    }
                                    local isHumidityValueAdded = false
                                    for _, otherFactorData in pairs(humidityFactors) do
                                        if otherFactorData.value == humidityFactorData.value then
                                            isHumidityValueAdded = true
                                            break
                                        end
                                    end
                                    if not isHumidityValueAdded then
                                        table.insert(humidityFactors, humidityFactorData)
                                    end
                                end
                            end)
                            if #humidityFactors > 0 then
                                table.sort(humidityFactors, function(pa, pb)
                                    return pa.value < pb.value
                                end)
                            end
                        end
                    end
                    if isDryingStateValid then
                        local soilTypeFactorsKey = pDryingStateKey .. ".soilTypeFactors"
                        if xmlFile:hasProperty(soilTypeFactorsKey) then
                            local soilTypeFactors = THUtils.clearTable(dryingFactors.soilTypes)
                            xmlFile:iterate(soilTypeFactorsKey .. ".soilType", function(_, pSoilTypeFactorKey)
                                local soilTypeIndex = THUtils.getXMLValue(xmlFile, pSoilTypeFactorKey, "#index")
                                local soilTypeFactor = THUtils.getXMLValue(xmlFile, pSoilTypeFactorKey, "#factor", 1)
                                local soilTypeInfo = pfManager:getSoilType(soilTypeIndex)
                                if soilTypeInfo == nil then
                                    THUtils.xmlLoadMsg(pSoilTypeFactorKey, "index", soilTypeIndex, false, THXMLMessage.INVALID_VALUE)
                                elseif soilTypeFactor < 0 then
                                    THUtils.xmlLoadMsg(pSoilTypeFactorKey, "factor", soilTypeFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                                else
                                    soilTypeFactors[soilTypeIndex] = soilTypeFactor
                                end
                            end)
                        end
                    end
                    if isDryingStateValid then
                        dryingStateData.index = #fruitTypeMoistureData.dryingStates + 1
                        fruitTypeMoistureData.dryingStates[dryingStateData.index] = dryingStateData
                    end
                end
            end)
        end
    end
    return loadSuccess
end
function THMoistureSystem:loadFromSavegameXMLFile(xmlFile, xmlKey)
    local loadSuccess = THMoistureSystem:superClass().loadFromSavegameXMLFile(self, xmlFile, xmlKey)
    if self.isServer and loadSuccess then
        local thFruitTypeManager = self.thFruitTypeManager
        self.dryingStateCachedTime = THUtils.getXMLValue(xmlFile, xmlKey, "#cachedDryingTime", 0)
        local dryingConditions = self.dryingConditions
        local dryingConditionsKey = xmlKey .. ".dryingConditions"
        if xmlFile:hasProperty(dryingConditionsKey) then
            dryingConditions.isRaining = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#isRaining", false)
            dryingConditions.isDayTime = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#isDayTime", false)
            dryingConditions.temperature = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#temp", 0)
            dryingConditions.dewPoint = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#dewPoint", 0)
            dryingConditions.periodLength = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#periodLength", 1)
            local humidity = THMoistureUtil.calculateHumidity(dryingConditions.temperature, dryingConditions.dewPoint)
            dryingConditions.humidity = MathUtil.clamp(humidity, dryingConditions.absMinHumidity, dryingConditions.absMaxHumidity)
            local currentRainData = dryingConditions.currentRainData
            local currentRainKey = dryingConditionsKey .. ".currentRain"
            if xmlFile:hasProperty(currentRainKey) then
                currentRainData.startDay = THUtils.getXMLValue(xmlFile, currentRainKey, "#startDay", 1)
                currentRainData.startTime = THUtils.getXMLValue(xmlFile, currentRainKey, "#startTime", 0)
                currentRainData.endDay = THUtils.getXMLValue(xmlFile, currentRainKey, "#endDay", 1)
                currentRainData.endTime = THUtils.getXMLValue(xmlFile, currentRainKey, "#endTime", 0)
                currentRainData.duration = THUtils.getXMLValue(xmlFile, currentRainKey, "#duration", 0)
            end
            local lastRainData = dryingConditions.lastRainData
            local lastRainKey = dryingConditionsKey .. ".lastRain"
            if xmlFile:hasProperty(lastRainKey) then
                lastRainData.startDay = THUtils.getXMLValue(xmlFile, lastRainKey, "#startDay", 1)
                lastRainData.startTime = THUtils.getXMLValue(xmlFile, lastRainKey, "#startTime", 0)
                lastRainData.endDay = THUtils.getXMLValue(xmlFile, lastRainKey, "#endDay", 1)
                lastRainData.endTime = THUtils.getXMLValue(xmlFile, lastRainKey, "#endTime", 0)
                lastRainData.duration = THUtils.getXMLValue(xmlFile, lastRainKey, "#duration", 0)
            end
            dryingConditions.isInitialized = true
        end
        xmlFile:iterate(xmlKey .. ".fruitTypes.fruitType", function(_, pFruitTypeKey)
            local fruitTypeName = THUtils.getXMLValue(xmlFile, pFruitTypeKey, "#name")
            local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitTypeName, true) -- force
            if fruitTypeMoistureData == nil then
                local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeName)
                if fruitTypeInfo == nil or fruitTypeInfo.index == FruitType.UNKNOWN then
                    THUtils.xmlLoadMsg(pFruitTypeKey, "name", fruitTypeName, false, THXMLMessage.INVALID_FRUIT_TYPE)
                else
                    THUtils.xmlLoadMsg(pFruitTypeKey, "name", fruitTypeName, true, "Could not load moisture values")
                end
            elseif fruitTypeMoistureData.isEnabled then
                xmlFile:iterate(pFruitTypeKey .. ".dryingStates.dryingState", function(_, pDryingStateKey)
                    local dryingStateIndex = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#index")
                    if dryingStateIndex ~= nil and fruitTypeMoistureData.dryingStates[dryingStateIndex] ~= nil then
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        local dryingTimes = dryingStateData.dryingTimes
                        local dryingFactors = dryingStateData.dryingFactors
                        local dryingTimesKey = pDryingStateKey .. ".dryingTimes"
                        dryingStateData.averageTemp = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#averageTemp", 0)
                        dryingStateData.averageWind = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#averageWind", 0)
                        dryingTimes.total = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#total", 0)
                        dryingTimes.night = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#night", 0)
                        dryingTimes.cloud = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#cloud", 0)
                        dryingTimes.rain = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#rain", 0)
                        dryingTimes.storm = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#storm", 0)
                        dryingTimes.snow = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#snow", 0)
                        local farmlandFactors = THUtils.clearTable(dryingFactors.farmlands)
                        xmlFile:iterate(pDryingStateKey .. ".farmlandFactors.farmland", function(_, pFarmlandFactorKey)
                            local farmlandId = THUtils.getXMLValue(xmlFile, pFarmlandFactorKey, "#id")
                            local farmlandFactor = THUtils.getXMLValue(xmlFile, pFarmlandFactorKey, "#factor")
                            if farmlandId ~= nil then
                                farmlandFactors[farmlandId] = farmlandFactor
                            end
                        end)
                    end
                end)
            end
        end)
    end
    return loadSuccess
end
function THMoistureSystem:saveToSavegameXMLFile(xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local saveSuccess = THMoistureSystem:superClass().saveToSavegameXMLFile(self, xmlFile, xmlKey)
    if self.isServer and saveSuccess then
        xmlFile:setValue(xmlKey .. "#cachedDryingTime", self.dryingStateCachedTime)
        local dryingConditions = self.dryingConditions
        local dryingConditionsKey = xmlKey .. ".dryingConditions"
        xmlFile:setValue(dryingConditionsKey .. "#isRaining", dryingConditions.isRaining)
        xmlFile:setValue(dryingConditionsKey .. "#isDayTime", dryingConditions.isDayTime)
        xmlFile:setValue(dryingConditionsKey .. "#temp", dryingConditions.temperature)
        xmlFile:setValue(dryingConditionsKey .. "#dewPoint", dryingConditions.dewPoint)
        xmlFile:setValue(dryingConditionsKey .. "#periodLength", dryingConditions.periodLength)
        local currentRainData = dryingConditions.currentRainData
        local lastRainData = dryingConditions.lastRainData
        if currentRainData.duration > 0 then
            local currentRainKey = dryingConditionsKey .. ".currentRain"
            xmlFile:setValue(currentRainKey .. "#startDay", currentRainData.startDay)
            xmlFile:setValue(currentRainKey .. "#startTime", currentRainData.startTime)
            xmlFile:setValue(currentRainKey .. "#endDay", currentRainData.endDay)
            xmlFile:setValue(currentRainKey .. "#endTime", currentRainData.endTime)
            xmlFile:setValue(currentRainKey .. "#duration", currentRainData.duration)
        end
        if currentRainData.duration > 0 then
            local lastRainKey = dryingConditionsKey .. ".lastRain"
            xmlFile:setValue(lastRainKey .. "#startDay", lastRainData.startDay)
            xmlFile:setValue(lastRainKey .. "#startTime", lastRainData.startTime)
            xmlFile:setValue(lastRainKey .. "#endDay", lastRainData.endDay)
            xmlFile:setValue(lastRainKey .. "#endTime", lastRainData.endTime)
            xmlFile:setValue(lastRainKey .. "#duration", lastRainData.duration)
        end
        local numFruitTypes = #self.fruitTypes
        if numFruitTypes > 0 then
            local fruitTypeXMLIndex = 0
            for fruitTypeIndex, fruitTypeMoistureData in ipairs(self.fruitTypes) do
                local fruitTypeKey = string.format("%s.fruitTypes.fruitType(%d)", xmlKey, fruitTypeXMLIndex)
                local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeIndex)
                if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= FruitType.UNKNOWN then
                    xmlFile:setValue(fruitTypeKey .. "#name", fruitTypeInfo.name)
                    local numDryingStates = #fruitTypeMoistureData.dryingStates
                    if numDryingStates > 0 then
                        local dryingStateXMLIndex = 0
                        local dryingStateKey = string.format("%s.dryingStates.dryingState(%d)", fruitTypeKey, dryingStateXMLIndex)
                        for dryingStateIndex, dryingStateData in ipairs(fruitTypeMoistureData.dryingStates) do
                            xmlFile:setValue(dryingStateKey .. "#index", dryingStateIndex)
                            xmlFile:setValue(dryingStateKey .. "#averageTemp", dryingStateData.averageTemp)
                            xmlFile:setValue(dryingStateKey .. "#averageWind", dryingStateData.averageWind)
                            local dryingTimesKey = dryingStateKey .. ".dryingTimes"
                            local dryingTimes = dryingStateData.dryingTimes
                            local dryingFactors = dryingStateData.dryingFactors
                            xmlFile:setValue(dryingTimesKey .. "#total", dryingTimes.total)
                            xmlFile:setValue(dryingTimesKey .. "#night", dryingTimes.night)
                            xmlFile:setValue(dryingTimesKey .. "#cloud", dryingTimes.cloud)
                            xmlFile:setValue(dryingTimesKey .. "#rain", dryingTimes.rain)
                            xmlFile:setValue(dryingTimesKey .. "#storm", dryingTimes.storm)
                            xmlFile:setValue(dryingTimesKey .. "#snow", dryingTimes.snow)
                            if next(dryingFactors.farmlands) ~= nil then
                                local farmlandXMLIndex = 0
                                for farmlandId, farmlandFactor in pairs(dryingFactors.farmlands) do
                                    local farmlandFactorKey = string.format("%s.farmlandFactors.farmland(%d)", dryingStateKey, farmlandXMLIndex)
                                    xmlFile:setValue(farmlandFactorKey .. "#id", farmlandId)
                                    xmlFile:setValue(farmlandFactorKey .. "#factor", farmlandFactor)
                                    farmlandXMLIndex = farmlandXMLIndex + 1
                                end
                            end
                            dryingStateXMLIndex = dryingStateXMLIndex + 1
                        end
                    end
                    fruitTypeXMLIndex = fruitTypeXMLIndex + 1
                end
            end
        end
    end
    return saveSuccess
end
function THMoistureSystem:onPostInit()
    if not self.isPostInitFinished then
        THMoistureSystem:superClass().onPostInit(self)
        self.isPostInitFinished = true
    end
end
function THMoistureSystem:onLoadEnvironment(environment, mission)
    THMoistureSystem:superClass().onLoadEnvironment(self, environment, mission)
    self.environment = environment
    self.growthSystem = mission.growthSystem
    self.weatherSystem = self.environment.weather
    self.windUpdater = self.weatherSystem.windUpdater
end
function THMoistureSystem:onLoadMapFinished(mission)
    THMoistureSystem:superClass().onLoadMapFinished(self, mission)
    self:initializeDataTables()
    self:loadConfiguration()
    self:finalizeDataTables()
    self:loadFromSavegame()
end
function THMoistureSystem:onUpdateTick(dt)
    THMoistureSystem:superClass().onUpdateTick(self, dt)
    if self.isServer then
        self.isDryingStateUpdateSafe = true
        if not self.dryingConditions.isInitialized then
            self:updateDryingConditions()
        end
        if not self.areDryingStatesInitialized then
            self:updateFruitTypeDryingStates(0)
            self.areDryingStatesInitialized = true
        end
    end
end
function THMoistureSystem:onReadStream(streamId, connection)
    if connection:getIsServer() then
        self:readMoistureFromStream(streamId, connection)
    end
end
function THMoistureSystem:onWriteStream(streamId, connection)
    if not connection:getIsServer() then
        self:writeMoistureToStream(streamId, connection, true)
    end
end
function THMoistureSystem:onReadUpdateStream(streamId, timeStamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isDirty
            self:readMoistureFromStream(streamId, connection)
        end
    end
end
function THMoistureSystem:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.dirtyFlag) ~= 0) then
            self:writeMoistureToStream(streamId, connection, false)
        end
    end
end
function THMoistureSystem:readMoistureFromStream(streamId, connection)
    if connection:getIsServer() then
        local farmlandManager = self.farmlandManager
        local dryingConditions = self.dryingConditions
        if streamReadBool(streamId) then -- dryingConditions initialized
            local humidity = streamReadUIntN(streamId, THMoistureSystem.FACTOR_SEND_BITS)
            local periodLength = streamReadUIntN(streamId, THMoistureSystem.PERIOD_SEND_BITS)
            dryingConditions.humidity = MathUtil.clamp(humidity / (2 ^ THMoistureSystem.FACTOR_SEND_BITS), 0, 1)
            dryingConditions.periodLength = periodLength
            dryingConditions.isInitialized = true
        end
        local numFruitTypes = #self.fruitTypes
        for i = 1, numFruitTypes do
            local fruitTypeMoistureData = self.fruitTypes[i]
            if streamReadBool(streamId) then -- needSync
                local fruitTypeIndex = fruitTypeMoistureData.fruitType.index
                local fruitTypeName = fruitTypeMoistureData.fruitType.name
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                THDebugUtil.printMsg(debugFlagId, nil, "ReadStream values for fruitType [%s] (%s):", fruitTypeIndex, fruitTypeName)
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        local dryingTimes = dryingStateData.dryingTimes
                        local dryingFactors = dryingStateData.dryingFactors
                        dryingTimes.total = streamReadFloat32(streamId)
                        THDebugUtil.printMsg(debugFlagId, nil, "- Drying times:")
                        THDebugUtil.printTable(debugFlagId, nil, dryingTimes)
                        if streamReadBool(streamId) then -- areFarmlandsDirty then
                            local farmlandFactors = THUtils.clearTable(dryingFactors.farmlands)
                            local numFarmlandBits = farmlandManager.numberOfBits or 32
                            local numFarmlandFactors = streamReadUIntN(streamId, numFarmlandBits)
                            if numFarmlandFactors > 0 then
                                for _ = 1, numFarmlandFactors do
                                    local farmlandId = streamReadUIntN(streamId, numFarmlandBits)
                                    local farmlandFactor = streamReadUIntN(streamId, 8)
                                    farmlandFactors[farmlandId] = farmlandFactor / 100
                                end
                            end
                            dryingFactors.areFarmlandsDirty = false
                            THDebugUtil.printMsg(debugFlagId, nil, "- Farmland factors:")
                            THDebugUtil.printTable(debugFlagId, nil, farmlandFactors)
                        end
                    end
                end
                fruitTypeMoistureData.needSync = false
            end
        end
    end
end
function THMoistureSystem:writeMoistureToStream(streamId, connection, isFullSync)
    if not connection:getIsServer() then
        local farmlandManager = self.farmlandManager
        local dryingConditions = self.dryingConditions
        if streamWriteBool(streamId, dryingConditions.isInitialized == true) then
            local humidity = MathUtil.clamp(dryingConditions.humidity, 0, 1) * (2 ^ THMoistureSystem.FACTOR_SEND_BITS)
            local periodLength = math.max(1, dryingConditions.periodLength)
            humidity = math.floor(humidity)
            streamWriteUIntN(streamId, humidity, THMoistureSystem.FACTOR_SEND_BITS)
            streamWriteUIntN(streamId, periodLength, THMoistureSystem.PERIOD_SEND_BITS)
        end
        local numFruitTypes = #self.fruitTypes
        for i = 1, numFruitTypes do
            local fruitTypeMoistureData = self.fruitTypes[i]
            if streamWriteBool(streamId, fruitTypeMoistureData.needSync == true or isFullSync == true) then
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        local dryingTimes = dryingStateData.dryingTimes
                        local dryingFactors = dryingStateData.dryingFactors
                        streamWriteFloat32(streamId, dryingTimes.total)
                        if streamWriteBool(streamId, dryingFactors.areFarmlandsDirty == true or isFullSync == true) then
                            local farmlandFactors = dryingFactors.farmlands
                            local numFarmlandFactors = THUtils.getTableLength(farmlandFactors)
                            local numFarmlandBits = farmlandManager.numberOfBits or 32
                            streamWriteUIntN(streamId, numFarmlandFactors, numFarmlandBits)
                            if numFarmlandFactors > 0 then
                                for farmlandId, farmlandFactor in pairs(farmlandFactors) do
                                    farmlandFactor = math.floor(farmlandFactor * 100)
                                    streamWriteUIntN(streamId, farmlandId, numFarmlandBits)
                                    streamWriteUIntN(streamId, farmlandFactor, 8)
                                end
                            end
                            dryingFactors.areFarmlandsDirty = false
                        end
                    end
                end
                fruitTypeMoistureData.needSync = false
            end
        end
    end
end
function THMoistureSystem:onMinuteChanged(currentMinute)
    if self.isServer then
        if not self.dryingConditions.isInitialized then
            self:updateDryingConditions()
        end
        self:updateFruitTypeDryingStates(1)
    end
end
function THMoistureSystem:onHourChanged(currentHour)
    if self.isServer then
        self:updateDryingConditions()
    end
end
function THMoistureSystem:onDayChanged(currentDay)
    if self.isServer then
        local numMoistureFruitTypes = #self.fruitTypes
        if numMoistureFruitTypes > 0 then
            for fruitTypeIndex = 1, numMoistureFruitTypes do
                local fruitTypeMoistureData = self.fruitTypes[fruitTypeIndex]
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        if not dryingStateData.hasResetPeriods then
                            self:resetFruitTypeDryingState(fruitTypeIndex, dryingStateIndex)
                        end
                    end
                end
            end
        end
    end
end
function THMoistureSystem:onPeriodChanged(currentPeriod)
    if self.isServer then
        local environment = self.environment
        local numMoistureFruitTypes = #self.fruitTypes
        self.dryingConditions.periodLength = THUtils.getNoNil(environment.daysPerPeriod, 1)
        if numMoistureFruitTypes > 0 then
            for fruitTypeIndex = 1, numMoistureFruitTypes do
                local fruitTypeMoistureData = self.fruitTypes[fruitTypeIndex]
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        if dryingStateData.hasResetPeriods
                            and dryingStateData.resetPeriods[currentPeriod]
                        then
                            self:resetFruitTypeDryingState(fruitTypeIndex, dryingStateIndex)
                        end
                    end
                end
            end
        end
        self:raiseDirtyFlags(self.dirtyFlag)
    end
end
function THMoistureSystem:onWeatherChanged()
    if self.isServer then
        self:updateDryingConditions()
    end
end
function THMoistureSystem:getCurrentDayTime()
    local weatherSystem = self.weatherSystem
    local currentTime = 0
    local currentDay = 1
    if weatherSystem ~= nil and weatherSystem.owner ~= nil then
        currentTime = THUtils.getNoNil(weatherSystem.owner.dayTime, 0)
        currentDay = THUtils.getNoNil(weatherSystem.owner.currentMonotonicDay, 1)
    end
    return currentTime, currentDay
end
function THMoistureSystem:getCurrentWeatherData()
    local weatherSystem = self.weatherSystem
    if weatherSystem ~= nil and weatherSystem.forecastItems ~= nil then
        local currentTime, currentDay = self:getCurrentDayTime()
        local numForecastItems = #weatherSystem.forecastItems
        if numForecastItems > 0 then
            for itemIndex = 1, numForecastItems do
                local forecastItem = weatherSystem.forecastItems[itemIndex]
                if forecastItem.startDay > currentDay
                    or (forecastItem.startDay == currentDay and forecastItem.startDayTime > currentTime)
                then
                    if itemIndex > 1 then
                        local lastForecastItem = weatherSystem.forecastItems[itemIndex - 1]
                        local weatherObject = weatherSystem:getWeatherObjectByIndex(lastForecastItem.season, lastForecastItem.objectIndex)
                        if weatherObject ~= nil then
                            return weatherObject.weatherType.index, weatherObject, lastForecastItem
                        end
                    end
                end
            end
        end
    end
    return WeatherType.SUN
end
function THMoistureSystem:getNextWeatherData()
    local weatherSystem = self.weatherSystem
    if weatherSystem ~= nil and weatherSystem.forecastItems ~= nil then
        local currentTime, currentDay = self:getCurrentDayTime()
        local numForecastItems = #weatherSystem.forecastItems
        if numForecastItems > 0 then
            for itemIndex = 1, numForecastItems do
                local forecastItem = weatherSystem.forecastItems[itemIndex]
                if forecastItem.startDay > currentDay
                    or (forecastItem.startDay == currentDay and forecastItem.startDayTime > currentTime)
                then
                    local weatherObject = weatherSystem:getWeatherObjectByIndex(forecastItem.season, forecastItem.objectIndex)
                    if weatherObject ~= nil then
                        return weatherObject.weatherType.index, weatherObject, forecastItem
                    end
                end
            end
        end
    end
    return WeatherType.SUN
end
function THMoistureSystem:getTimeSinceRain()
    local weatherSystem = self.weatherSystem
    local timeSinceRain = 0
    if weatherSystem ~= nil and weatherSystem.owner ~= nil then
        if not weatherSystem:getIsRaining() then
            timeSinceRain = weatherSystem:getTimeSinceLastRain()
            if timeSinceRain == nil or timeSinceRain < 0 or timeSinceRain == math.huge then
                local currentTime, currentDay = self:getCurrentDayTime()
                local lastRainData = self.dryingConditions.lastRainData
                if lastRainData.endDay > currentDay
                    or (lastRainData.endDay == currentDay and lastRainData.endTime > currentTime)
                then
                    timeSinceRain = 0
                else
                    local absCurrentTime = currentTime + ((currentDay - lastRainData.endDay) * weatherSystem.owner.dayLength)
                    timeSinceRain = (absCurrentTime - lastRainData.endTime) / 60000 -- ms to minutes
                end
            end
        end
    end
    return timeSinceRain
end
function THMoistureSystem:getCurrentHumidityValues()
    local weatherSystem = self.weatherSystem
    local dryingConditions = self.dryingConditions
    if dryingConditions.isInitialized then
        return dryingConditions.humidity, dryingConditions.dewPoint
    else
        local currentTemp = weatherSystem:getCurrentTemperature()
        local absMinHumidity = dryingConditions.absMinHumidity
        local absMaxHumidity = dryingConditions.absMaxHumidity
        local humidity = (absMinHumidity + absMaxHumidity) / 2
        local dewPoint = THMoistureUtil.calculateDewPoint(currentTemp, humidity)
        return humidity, dewPoint
    end
end
function THMoistureSystem:getTimeUntilRain(isAbsolute)
    local weatherSystem = self.weatherSystem
    local timeUntilRain = math.huge
    if weatherSystem ~= nil and weatherSystem.owner ~= nil and weatherSystem.forecastItems ~= nil then
        if weatherSystem:getIsRaining() then
            timeUntilRain = 0
        else
            local currentTime, currentDay = self:getCurrentDayTime()
            local dayLength = weatherSystem.owner.dayLength
            local numForecastItems = #weatherSystem.forecastItems
            if numForecastItems > 0 then
                for itemIndex = 1, numForecastItems do -- start at next forecast item
                    local forecastItem = weatherSystem.forecastItems[itemIndex]
                    local weatherObject = weatherSystem:getWeatherObjectByIndex(forecastItem.season, forecastItem.objectIndex)
                    if weatherObject ~= nil then
                        if forecastItem.startDay > currentDay
                            or (forecastItem.startDay == currentDay and forecastItem.startDayTime > currentTime)
                        then
                            if weatherObject.weatherType.index == WeatherType.RAIN
                                or weatherObject.weatherType.index == WeatherType.SNOW
                            then
                                local absStartTime = forecastItem.startDayTime + ((forecastItem.startDay - currentDay) * dayLength)
                                timeUntilRain = (absStartTime - currentTime) / 60000 -- ms to minutes
                                break
                            elseif not isAbsolute then -- stop after first valid attempt
                                break
                            end
                        end
                    end
                end
            end
        end
    end
    return timeUntilRain
end
function THMoistureSystem:updateDryingConditions(noEventSend)
    if self.isServer then
        local environment = self.environment
        local weatherSystem = self.weatherSystem
        local dryingConditions = self.dryingConditions
        if environment ~= nil and weatherSystem ~= nil and weatherSystem.owner ~= nil then
            local currentTime, currentDay = self:getCurrentDayTime()
            local dayLength = weatherSystem.owner.dayLength
            local periodLength = dryingConditions.periodLength
            local curTemp = weatherSystem:getCurrentTemperature()
            local isDayTime = environment.isSunOn
            local isRaining = weatherSystem:getIsRaining()
            local weatherType, _, forecastItem = self:getCurrentWeatherData()
            local nextWeatherType, _, nextForecastItem = self:getNextWeatherData()
            local absMinHumidity = dryingConditions.absMinHumidity
            local absMaxHumidity = dryingConditions.absMaxHumidity
            local absMinDewPoint = THMoistureUtil.calculateDewPoint(curTemp, absMinHumidity)
            local absMaxDewPoint = THMoistureUtil.calculateDewPoint(curTemp, absMaxHumidity)
            absMinDewPoint = math.max(absMinDewPoint, -33)
            absMaxDewPoint = math.min(absMaxDewPoint, 35)
            local timeSinceRain = self:getTimeSinceRain()
            local timeUntilRain = self:getTimeUntilRain(true)
            local currentRainData = dryingConditions.currentRainData
            local lastRainData = dryingConditions.lastRainData
            if not dryingConditions.isInitialized then
                dryingConditions.periodLength = THUtils.getNoNil(environment.daysPerPeriod, 1)
                dryingConditions.temperature = curTemp
                dryingConditions.isDayTime = isDayTime
                dryingConditions.isRaining = isRaining
                if isRaining then
                    dryingConditions.humidity = 1
                else
                    dryingConditions.humidity = 0.5
                end
                dryingConditions.dewPoint = THMoistureUtil.calculateDewPoint(dryingConditions.temperature, dryingConditions.humidity)
                lastRainData.startDay = currentDay
                lastRainData.startTime = currentTime
                lastRainData.endDay = currentDay
                lastRainData.endTime = currentTime
                lastRainData.duration = 0
                for key, value in pairs(lastRainData) do
                    currentRainData[key] = value
                end
                dryingConditions.isInitialized = true
            end
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THDebugUtil.printMsg(debugFlagId, nil, "Current drying conditions:")
                THDebugUtil.printMsg(debugFlagId, nil, "Current temperature: %s", curTemp)
                THDebugUtil.printMsg(debugFlagId, nil, "Abs max dew point: %s", absMaxDewPoint)
                THDebugUtil.printMsg(debugFlagId, nil, "Abs min dew point: %s", absMinDewPoint)
                THDebugUtil.printMsg(debugFlagId, nil, "")
            end
            if isRaining or timeSinceRain <= 0 or timeUntilRain <= 0
                or weatherType == WeatherType.RAIN
                or weatherType == WeatherType.SNOW
            then
                if isRaining then -- raining flag specifically set
                    currentRainData.endTime = currentTime
                    currentRainData.endDay = currentDay
                    if isRaining ~= dryingConditions.isRaining then -- new rain event
                        currentRainData.startDay = currentDay
                        currentRainData.startTime = currentTime
                        currentRainData.duration = 0 -- duration calculated at end of event
                    end
                end
                dryingConditions.dewPoint = absMaxDewPoint
            else
                if isRaining ~= dryingConditions.isRaining then -- rain event ended
                    local absEndTime = currentTime
                    if currentDay > currentRainData.startDay then
                        absEndTime = currentRainData.endTime + ((currentDay - currentRainData.startDay) * dayLength)
                    end
                    currentRainData.duration = absEndTime - currentRainData.startTime
                    for key, value in pairs(currentRainData) do
                        lastRainData[key] = value
                    end
                end
                local lastTemp = dryingConditions.temperature
                local tempTrend = curTemp - lastTemp
                local weatherFactor = 0
                local rainFactor = 0
                local dryingFactor = dryingConditions.dryingFactor
                if timeSinceRain < 0 or timeUntilRain < 0 or dryingFactor <= 0 then
                    rainFactor = 1
                else
                    local maxDryingTime = (dryingConditions.maxDryingTime * periodLength) / dryingFactor
                    if timeSinceRain < math.huge then
                        if timeSinceRain < maxDryingTime then
                            rainFactor = 1 - MathUtil.clamp(timeSinceRain / maxDryingTime, 0, 1)
                        end
                    end
                    if timeUntilRain < math.huge then
                        if timeUntilRain < maxDryingTime then
                            local untilRainFactor = 1 - MathUtil.clamp(timeUntilRain / maxDryingTime, 0, 1)
                            rainFactor = math.max(rainFactor, untilRainFactor)
                        end
                    end
                end
                if forecastItem ~= nil and nextForecastItem ~= nil then
                    local function getWeatherTypeFactor(pWeatherType)
                        local factor = 0 -- clear sunny weather
                        if pWeatherType ~= WeatherType.SUN then
                            if pWeatherType == WeatherType.RAIN
                                or pWeatherType == WeatherType.SNOW
                            then
                                factor = 1
                            else
                                factor = 0.5
                            end
                        end
                        return factor
                    end
                    local weatherTypeFactor = getWeatherTypeFactor(weatherType)
                    local nextWeatherTypeFactor = getWeatherTypeFactor(nextWeatherType)
                    local weatherStartDay = forecastItem.startDay
                    local weatherStartTime = forecastItem.startDayTime
                    local weatherEndDay = nextForecastItem.startDay
                    local weatherEndTime = nextForecastItem.startDayTime
                    local absCurrentTime = currentTime
                    local absWeatherEndTime = weatherEndTime
                    if currentDay > weatherStartDay then
                        absCurrentTime = currentTime + ((currentDay - weatherStartDay) * dayLength)
                    end
                    if weatherEndDay > weatherStartDay then
                        absWeatherEndTime = weatherEndTime + ((weatherEndDay - weatherStartDay) * dayLength)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "Weather start time: %s", weatherStartTime)
                    THDebugUtil.printMsg(debugFlagId, nil, "Weather end time: %s", absWeatherEndTime)
                    THDebugUtil.printMsg(debugFlagId, nil, "Current weather time: %s", absCurrentTime)
                    if absCurrentTime >= weatherStartTime and absWeatherEndTime >= absCurrentTime then
                        local weatherTimeAlpha = THUtils.getFactorInRange(weatherStartTime, absWeatherEndTime, absCurrentTime)
                        THDebugUtil.printMsg(debugFlagId, nil, "Current weather alpha: %s", weatherTimeAlpha)
                        weatherFactor = MathUtil.clamp(weatherTypeFactor + ((nextWeatherTypeFactor - weatherTypeFactor) * weatherTimeAlpha), 0, 1)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                end
                local moistureFactor = (weatherFactor + rainFactor) / 2  -- clamped (0 - 1) above
                local lastDewPoint = dryingConditions.dewPoint
                local nextDewPoint = lastDewPoint + (tempTrend * 0.75 * math.random())
                local minDewPoint = math.min(absMaxDewPoint, absMinDewPoint + ((absMaxDewPoint - absMinDewPoint) * moistureFactor))
                local maxDewPoint = absMaxDewPoint
                if tempTrend > 0 then
                    nextDewPoint = lastDewPoint + (tempTrend * 0.75 * moistureFactor * math.random())
                elseif tempTrend < 0 then
                    nextDewPoint = lastDewPoint + (tempTrend * 0.75 * (1 - moistureFactor) * math.random())
                end
                if isRaining ~= dryingConditions.isRaining -- implies rain event ended (isRaining handled above)
                    or (nextDewPoint >= maxDewPoint and lastDewPoint >= maxDewPoint)
                then
                    nextDewPoint = maxDewPoint - ((maxDewPoint - minDewPoint) * (1 - moistureFactor) * math.random())
                elseif nextDewPoint <= minDewPoint and lastDewPoint <= minDewPoint then
                    nextDewPoint = minDewPoint + ((maxDewPoint - minDewPoint) * moistureFactor * math.random())
                end
                dryingConditions.dewPoint = MathUtil.clamp(nextDewPoint, minDewPoint, maxDewPoint)
                THDebugUtil.printMsg(debugFlagId, nil, "Current weather type: %s", weatherType)
                THDebugUtil.printMsg(debugFlagId, nil, "Next weather type: %s", nextWeatherType)
                THDebugUtil.printMsg(debugFlagId, nil, "Weather factor : %s", weatherFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "")
                THDebugUtil.printMsg(debugFlagId, nil, "Last temperature: %s", lastTemp)
                THDebugUtil.printMsg(debugFlagId, nil, "Last dew point: %s", lastDewPoint)
                THDebugUtil.printMsg(debugFlagId, nil, "Adjusted min dew point: %s", minDewPoint)
                THDebugUtil.printMsg(debugFlagId, nil, "")
                THDebugUtil.printMsg(debugFlagId, nil, "Time since rain: %s", timeSinceRain)
                THDebugUtil.printMsg(debugFlagId, nil, "Time until rain: %s", timeUntilRain)
                THDebugUtil.printMsg(debugFlagId, nil, "Rain factor: %s", rainFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "")
            end
            local nextHumidity = THMoistureUtil.calculateHumidity(curTemp, dryingConditions.dewPoint)
            dryingConditions.temperature = curTemp
            dryingConditions.isDayTime = isDayTime
            dryingConditions.isRaining = isRaining
            dryingConditions.humidity = MathUtil.clamp(nextHumidity, absMinHumidity, absMaxHumidity)
            THDebugUtil.printMsg(debugFlagId, nil, "Current Humidity: %0.1f%%", dryingConditions.humidity * 100)
            THDebugUtil.printTable(debugFlagId, nil, dryingConditions, 1)
            if noEventSend ~= true then
                self:raiseDirtyFlags(self.dirtyFlag)
            end
        end
    end
end
function THMoistureSystem:createFruitTypeDryingState(fruitType, minStateId, maxStateId, pctStart, pctEnd)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    if fruitTypeMoistureData ~= nil then
        local fruitTypeData = fruitTypeMoistureData.fruitType
        local fruitTypeIndex = fruitTypeData.index
        local dryingStateData = {}
        dryingStateData.parent = fruitTypeMoistureData
        local minGrowthState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, minStateId, true)
        local maxGrowthState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, maxStateId, true)
        if minGrowthState == nil or minGrowthState < 0 then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "minStateId", minStateId)
        elseif maxGrowthState == nil or maxGrowthState < 0 then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "maxStateId", maxStateId)
        elseif pctStart == nil or pctStart < 0 or pctStart > 1 then
            THUtils.stackMsg("Argument pctStart (%s) must be between 0 and 1", pctStart)
        elseif pctEnd == nil or pctEnd < 0 or pctEnd > 1 then
            THUtils.stackMsg("Argument pctEnd (%s) must be between 0 and 1", pctEnd)
        else
            dryingStateData.minGrowthState = math.min(minGrowthState, maxGrowthState)
            dryingStateData.maxGrowthState = math.max(minGrowthState, maxGrowthState)
            dryingStateData.timeToDry = 360 -- 6 hours to dry
            dryingStateData.averageWind = 0
            dryingStateData.averageTemp = 0
            dryingStateData.startPercent = pctStart
            dryingStateData.endPercent = pctEnd
            dryingStateData.dryingTimes = {
                total = 0,
                night = 0,
                cloud = 0,
                rain = 0,
                storm = 0,
                snow = 0
            }
            dryingStateData.dryingFactors = {
                night = THMoistureSystem.DEFAULT_DRYING_FACTORS.NIGHT,
                cloud = THMoistureSystem.DEFAULT_DRYING_FACTORS.CLOUD,
                rain = THMoistureSystem.DEFAULT_DRYING_FACTORS.RAIN,
                storm = THMoistureSystem.DEFAULT_DRYING_FACTORS.STORM,
                snow = THMoistureSystem.DEFAULT_DRYING_FACTORS.SNOW,
                wind = THMoistureSystem.DEFAULT_DRYING_FACTORS.WIND,
                humidity = {},
                soilTypes = {},
                farmlands = {},
                areFarmlandsDirty = false
            }
            local humidityFactors = dryingStateData.dryingFactors.humidity
            local absMinHumidity = self.dryingConditions.absMinHumidity
            local absMaxHumidity = self.dryingConditions.absMaxHumidity
            table.insert(humidityFactors, { value = absMinHumidity, factor = 1 })
            table.insert(humidityFactors, { value = absMaxHumidity, factor = 0 })
            dryingStateData.hasResetPeriods = true
            dryingStateData.resetPeriods = {}
            return dryingStateData
        end
    end
end
function THMoistureSystem:resetFruitTypeDryingState(fruitType, dryingStateIndex, noEventSend)
    local farmlandManager = self.farmlandManager
    local weatherSystem = self.weatherSystem
    local windUpdater = self.windUpdater
    if self.isServer or noEventSend == true then
        local dryingStateData = self:getFruitTypeDryingState(fruitType, dryingStateIndex)
        if dryingStateData == nil then
            return false
        end
        local fruitTypeMoistureData = dryingStateData.parent
        local minTemp, maxTemp = weatherSystem:getCurrentMinMaxTemperatures()
        local _, _, windSpeed = windUpdater:getCurrentValues()
        dryingStateData.averageTemp = (minTemp + maxTemp) / 2
        dryingStateData.averageWind = windSpeed
        local dryingTimes = dryingStateData.dryingTimes
        local dryingFactors = dryingStateData.dryingFactors
        dryingTimes.total = 0
        dryingTimes.night = 0
        dryingTimes.cloud = 0
        dryingTimes.rain = 0
        dryingTimes.storm = 0
        dryingTimes.snow = 0
        THUtils.clearTable(dryingFactors.farmlands)
        if self.isServer then
            local farmlandsList = farmlandManager:getFarmlands()
            for _, farmlandInfo in pairs(farmlandsList) do
                if farmlandInfo.id ~= nil and farmlandInfo.id > 0 then
                    dryingFactors.farmlands[farmlandInfo.id] = math.floor(math.random() * 100) / 100
                    dryingFactors.areFarmlandsDirty = true
                end
            end
            fruitTypeMoistureData.needSync = true
            if noEventSend ~= true then
                self:raiseDirtyFlags(self.dirtyFlag)
            end
        end
        return true
    end
    return false
end
function THMoistureSystem:updateFruitTypeDryingStates(elapsedTime)
    local updateSuccess = true
    if self.isServer then
        if self.isDryingStateUpdateSafe then
            local numMoistureFruitTypes = #self.fruitTypes
            elapsedTime = elapsedTime + self.dryingStateCachedTime
            self.dryingStateCachedTime = 0
            if numMoistureFruitTypes > 0 then
                for fruitTypeIndex = 1, numMoistureFruitTypes do
                    local fruitTypeMoistureData = self.fruitTypes[fruitTypeIndex]
                    local numDryingStates = #fruitTypeMoistureData.dryingStates
                    if numDryingStates > 0 then
                        for dryingStateIndex = 1, numDryingStates do
                            if not self:updateFruitTypeDryingState(fruitTypeIndex, dryingStateIndex, elapsedTime) then
                                updateSuccess = false
                            end
                        end
                    end
                end
            end
        else
            self.dryingStateCachedTime = self.dryingStateCachedTime + elapsedTime
        end
        self.isDryingStateUpdateSafe = false
        self:raiseActive()
    else
        updateSuccess = false
    end
    return updateSuccess
end
function THMoistureSystem:updateFruitTypeDryingState(fruitType, dryingStateIndex, elapsedTime, noEventSend)
    local farmlandManager = self.farmlandManager
    local thFruitTypeManager = self.thFruitTypeManager
    local environment = self.environment
    local weatherSystem = self.weatherSystem
    local windUpdater = self.windUpdater
    local updateSuccess = true
    if self.isServer then
        local dryingConditions = self.dryingConditions
        local dryingStateData = self:getFruitTypeDryingState(fruitType, dryingStateIndex)
        if dryingStateData == nil then
            local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
            if fruitTypeInfo == nil then
                THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "fruitType", fruitType)
                updateSuccess = false
            elseif THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.ERROR) then
                THDebugUtil.printMsg(debugFlagId, nil, "No drying state information found for fruitType (%s)", fruitTypeInfo.name)
            end
        else
            local fruitTypeMoistureData = dryingStateData.parent
            local dryingTimes = dryingStateData.dryingTimes
            local dryingFactors = dryingStateData.dryingFactors
            local weatherType = weatherSystem:getCurrentWeatherType()
            local isDayTime = environment.isSunOn
            local curTemp = weatherSystem:getCurrentTemperature()
            local curHumidity = dryingConditions.humidity
            local _, _, windSpeed = windUpdater:getCurrentValues()
            local windFactor = MathUtil.clamp(windSpeed / WindUpdater.MAX_SPEED, 0, 1) * dryingFactors.wind
            local humidityFactor = self:getDryingStateHumidityFactor(dryingStateData, curHumidity)
            local weatherFactor = 1
            if weatherType == WeatherType.CLOUDY then
                weatherFactor = dryingFactors.cloud
            elseif weatherType == WeatherType.RAIN then
                weatherFactor = dryingFactors.rain
            elseif weatherType == WeatherType.SNOW then
                weatherFactor = dryingFactors.snow
            end
            if not isDayTime then
                weatherFactor = math.min(weatherFactor, dryingFactors.night)
            end
            local adjustedHumidityFactor = (humidityFactor + windFactor) / 2
            local totalDryingFactor = math.min(adjustedHumidityFactor, weatherFactor)
            local adjustedTime = elapsedTime * totalDryingFactor -- can be negative
            dryingTimes.total = math.max(0, dryingTimes.total + adjustedTime)
            if not isDayTime then
                dryingTimes.night = math.max(0, dryingTimes.night + adjustedTime)
            end
            if weatherType == WeatherType.CLOUDY then
                dryingTimes.cloud = math.max(0, dryingTimes.cloud + adjustedTime)
            elseif weatherType == WeatherType.RAIN then
                dryingTimes.rain = math.max(0, dryingTimes.rain + adjustedTime)
            elseif weatherType == WeatherType.SNOW then
                dryingTimes.snow = math.max(0, dryingTimes.snow + adjustedTime)
            end
            dryingStateData.averageTemp = (dryingStateData.averageTemp + curTemp) / 2
            dryingStateData.averageWind = (dryingStateData.averageWind + windSpeed) / 2
            if next(dryingFactors.farmlands) == nil then
                local farmlandsList = farmlandManager:getFarmlands()
                if farmlandsList ~= nil and next(farmlandsList) ~= nil then
                    for _, farmlandInfo in pairs(farmlandsList) do
                        if farmlandInfo.id ~= nil and farmlandInfo.id > 0 then
                            dryingFactors.farmlands[farmlandInfo.id] = math.floor(math.random() * 100) / 100
                            dryingFactors.areFarmlandsDirty = true
                        end
                    end
                end
            end
            if THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.UPDATE) then
                local fruitTypeName = fruitTypeMoistureData.fruitType.name
                THDebugUtil.printMsg(debugFlagId, nil, "Updated drying state [%s] for fruitType %s", dryingStateIndex, fruitTypeName)
                THDebugUtil.printMsg(debugFlagId, nil, "- Wind speed: %s", windSpeed)
                THDebugUtil.printMsg(debugFlagId, nil, "- Wind factor: %s", windFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "- Humidity: %s", curHumidity)
                THDebugUtil.printMsg(debugFlagId, nil, "- Humidity factor: %s", humidityFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "- Weather factor: %s", weatherFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "- Total drying factor: %s", totalDryingFactor)
                THDebugUtil.printTable(debugFlagId, nil, dryingStateData, 1)
            end
            fruitTypeMoistureData.needSync = true
            if noEventSend ~= true then
                self:raiseDirtyFlags(self.dirtyFlag)
            end
        end
    else
        updateSuccess = false
    end
    return updateSuccess
end
function THMoistureSystem:getFruitTypeMoistureData(fruitType, force)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
    local fruitTypeMoistureData = nil
    if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= FruitType.UNKNOWN then
        fruitTypeMoistureData = self.fruitTypes[fruitTypeInfo.index]
        if fruitTypeMoistureData ~= nil and force ~= true then
            if not fruitTypeMoistureData.isEnabled then
                fruitTypeMoistureData = nil
            end
        end
    end
    return fruitTypeMoistureData
end
function THMoistureSystem:getFruitTypeDryingState(fruitType, dryingStateIndex)
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    if fruitTypeMoistureData ~= nil and dryingStateIndex ~= nil then
        return fruitTypeMoistureData.dryingStates[dryingStateIndex], fruitTypeMoistureData
    end
end
function THMoistureSystem:getFruitTypeDryingStateByGrowthState(fruitType, growthStateId)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    local growthState = thFruitTypeManager:getFruitTypeGrowthState(fruitType, growthStateId, true)
    if growthState == nil or growthState < 0 then
        THDebugUtil.printMsg(debugFlagId, nil, THMessage.ARGUMENT_INVALID, "growthStateId", growthStateId)
    elseif fruitTypeMoistureData ~= nil then
        local dryingStateIndex = fruitTypeMoistureData.dryingStateMapping[growthState]
        if dryingStateIndex ~= nil then
            return fruitTypeMoistureData.dryingStates[dryingStateIndex], fruitTypeMoistureData
        end
    end
end
function THMoistureSystem:getFruitTypeMoisture(fruitType, growthStateId, harvestMultiplier, farmlandId, soilTypeIndex)
    local thFruitTypeManager = self.thFruitTypeManager
    local pfManager = self.thPFManager
    local dryingConditions = self.dryingConditions
    local periodLength = dryingConditions.periodLength
    local dryingStateData, fruitTypeMoistureData = self:getFruitTypeDryingStateByGrowthState(fruitType, growthStateId)
    local currentMoisture = -1
    local dryFactor = -1
    if dryingStateData ~= nil and fruitTypeMoistureData ~= nil then
        local fruitTypeData = fruitTypeMoistureData.fruitType.data
        local startMoisture = dryingStateData.startPercent
        local endMoisture = dryingStateData.endPercent
        if not dryingConditions.isInitialized then
            local growthState = thFruitTypeManager:getFruitTypeGrowthState(fruitType, growthStateId, true)
            if fruitTypeData ~= nil and growthState ~= nil and growthState >= 0 then
                local minHarvestState = fruitTypeData.minHarvestState or -1
                local maxHarvestState = fruitTypeData.maxHarvestState or -1
                if maxHarvestState > 0 and growthState > maxHarvestState then
                    dryFactor = 1
                elseif minHarvestState > 0 then
                    dryFactor = 0
                end
            end
        else
            local dryingTimes = dryingStateData.dryingTimes
            local timeToDry = dryingStateData.timeToDry * periodLength
            local totalDryingTime = dryingTimes.total
            if timeToDry <= 0 then
                if totalDryingTime > 0 then
                    dryFactor = 1
                else
                    dryFactor = 0
                end
            else
                local yieldDryingFactor = 1
                local farmlandDryingFactor = 1
                local soilTypeDryingFactor = 1
                if harvestMultiplier ~= nil then
                    yieldDryingFactor = 1 - THUtils.getFactorInRange(1, 2, harvestMultiplier)
                end
                if farmlandId ~= nil then
                    farmlandDryingFactor = self:getDryingStateFarmlandFactor(dryingStateData, farmlandId)
                end
                local soilTypeInfo = pfManager:getSoilType(soilTypeIndex)
                local soilTypeFactor = self:getDryingStateSoilTypeFactor(dryingStateData, soilTypeIndex)
                if soilTypeInfo ~= nil then
                    local minPotential = pfManager.minSoilTypeYieldPotential
                    local maxPotential = pfManager.maxSoilTypeYieldPotential
                    local curPotential = soilTypeInfo.yieldPotential or maxPotential
                    soilTypeDryingFactor = THUtils.getFactorInRange(minPotential, maxPotential, curPotential)
                end
                THDebugUtil.printMsg(debugFlagId, nil, "Current moisture for fruitType: %s", fruitTypeMoistureData.fruitType.name)
                THDebugUtil.printMsg(debugFlagId, nil, "- Drying time: %s", totalDryingTime)
                THDebugUtil.printMsg(debugFlagId, nil, "- Time to dry: %s", timeToDry)
                THDebugUtil.printMsg(debugFlagId, nil, "- Yield drying factor: %s", yieldDryingFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "- Farmland drying factor: %s", farmlandDryingFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "- SoilType drying factor: %s", soilTypeDryingFactor)
                timeToDry = (timeToDry * 0.5) + math.max(0, (timeToDry * 0.5 * (1 - farmlandDryingFactor)))
                timeToDry = timeToDry + math.max(0, (timeToDry * (1 - soilTypeDryingFactor)))
                totalDryingTime = math.max(0, totalDryingTime * soilTypeFactor)
                if totalDryingTime < timeToDry then
                    dryFactor = MathUtil.clamp(totalDryingTime / timeToDry, 0, 1)
                else
                    dryFactor = 1
                end
                THDebugUtil.printMsg(debugFlagId, nil, "- Drying time (adjusted): %s", totalDryingTime)
                THDebugUtil.printMsg(debugFlagId, nil, "- Time to dry (adjusted): %s", timeToDry)
                THDebugUtil.printMsg(debugFlagId, nil, "- Dry factor: %s", dryFactor)
                THDebugUtil.printMsg(debugFlagId, nil, "")
            end
        end
        if dryFactor >= 0 then
            currentMoisture = startMoisture + ((endMoisture - startMoisture) * dryFactor)
            THDebugUtil.printMsg(debugFlagId, nil, "- Starting moisture: %s", startMoisture)
            THDebugUtil.printMsg(debugFlagId, nil, "- Ending moisture: %s", endMoisture)
            THDebugUtil.printMsg(debugFlagId, nil, "- Current moisture: %s", currentMoisture)
            THDebugUtil.printMsg(debugFlagId, nil, "")
        end
    end
    return currentMoisture, dryFactor
end
function THMoistureSystem:getDryingStateHumidityFactor(dryingStateData, humidity)
    local humidityFactors = dryingStateData.dryingFactors.humidity
    local numHumidityFactors = #humidityFactors
    local humidityFactor = 1
    if numHumidityFactors > 0 then
        for factorIndex = 1, numHumidityFactors do
            local factorEntry = humidityFactors[factorIndex]
            if humidity <= factorEntry.value then
                if factorIndex == 1 then
                    humidityFactor = factorEntry.factor
                else
                    local lastFactorEntry = humidityFactors[factorIndex - 1]
                    local humidityAlpha = THUtils.getFactorInRange(lastFactorEntry.value, factorEntry.value, humidity)
                    humidityFactor = THUtils.getValueInRange(lastFactorEntry.factor, factorEntry.factor, humidityAlpha)
                end
                break
            end
        end
    end
    return humidityFactor
end
function THMoistureSystem:getDryingStateSoilTypeFactor(dryingStateData, soilTypeIndex)
    local soilTypeFactors = dryingStateData.dryingFactors.soilTypes
    local soilTypeFactor = 1
    if soilTypeIndex ~= nil and soilTypeIndex > 0 then
        soilTypeFactor = soilTypeFactors[soilTypeIndex] or 1
    end
    return soilTypeFactor
end
function THMoistureSystem:getDryingStateFarmlandFactor(dryingStateData, farmlandId)
    local farmlandFactors = dryingStateData.dryingFactors.farmlands
    local farmlandFactor = 1
    if farmlandId ~= nil and farmlandId > 0 then
        farmlandFactor = farmlandFactors[farmlandId] or 1
    end
    return farmlandFactor
end
function THMoistureSystem:setFillTypeMoistureFactors(fillType, moisture, priceFactor, yieldFactor, decayFactor)
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType, true) -- force
    if THUtils.argIsValid(fillTypeMoistureData ~= nil, "fillType", fillType)
        and THUtils.argIsValid(type(moisture) == "number" and moisture >= 0, "moisture", moisture)
        and THUtils.argIsValid(type(priceFactor) == "number" and priceFactor >= 0, "priceFactor", priceFactor)
        and THUtils.argIsValid(type(yieldFactor) == "number" and yieldFactor >= 0, "yieldFactor", yieldFactor)
        and THUtils.argIsValid(type(decayFactor) == "number" and decayFactor >= 0 and decayFactor <= 1, "decayFactor", decayFactor)
        and fillTypeMoistureData.isEnabled
    then
        local curveEntries = fillTypeMoistureData.factors.curve
        local factorData = nil
        for _, otherFactorData in pairs(curveEntries) do
            if moisture == otherFactorData.moisture then
                factorData = otherFactorData
                break
            end
        end
        if factorData == nil then
            factorData = {
                moisture = moisture,
                priceFactor = priceFactor,
                yieldFactor = yieldFactor,
                decayFactor = decayFactor
            }
            table.insert(curveEntries, factorData)
            table.sort(curveEntries, function(pa, pb)
                return pa.moisture < pb.moisture
            end)
        else
            factorData.priceFactor = priceFactor
            factorData.yieldFactor = yieldFactor
            factorData.decayFactor = decayFactor
        end
        return true
    end
    return false
end
function THMoistureSystem:getFillTypeMoistureData(fillType, force)
    local thFillTypeManager = self.thFillTypeManager
    local fillTypeInfo = thFillTypeManager:getFillType(fillType)
    local fillTypeMoistureData = nil
    if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
        fillTypeMoistureData = self.fillTypes[fillTypeInfo.index]
        if fillTypeMoistureData ~= nil and force ~= true then
            if not fillTypeMoistureData.isEnabled
                or not fillTypeMoistureData.hasMoisture
            then
                fillTypeMoistureData = nil
            end
        end
    end
    return fillTypeMoistureData
end
function THMoistureSystem:getFillTypeMoisture(fruitType, fillType, growthStateId, harvestMultiplier, farmlandId, soilTypeIndex)
    local fruitTypeMoisture = self:getFruitTypeMoisture(fruitType, growthStateId, harvestMultiplier, farmlandId, soilTypeIndex)
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType)
    local fillTypeMoisture = -1
    if fillTypeMoistureData ~= nil and fruitTypeMoisture >= 0 then
        local minMoisture = fillTypeMoistureData.minPercent
        local maxMoisture = fillTypeMoistureData.maxPercent
        fillTypeMoisture = MathUtil.clamp(fruitTypeMoisture, minMoisture, maxMoisture)
    end
    return fillTypeMoisture
end
function THMoistureSystem:getFillTypeMoistureFactors(fillType, moisture)
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType)
    local priceFactor = 1
    local yieldFactor = 1
    local decayFactor = 0
    if fillTypeMoistureData ~= nil then
        local moistureFactors = fillTypeMoistureData.factors
        local factorsCurve = moistureFactors.curve
        local numCurveEntries = #factorsCurve
        if numCurveEntries > 0 then
            moisture = MathUtil.clamp(moisture, fillTypeMoistureData.minPercent, fillTypeMoistureData.maxPercent)
            for curveIndex = 1, numCurveEntries do
                local curveEntry = factorsCurve[curveIndex]
                if moisture <= curveEntry.moisture then
                    local lastCurveEntry = factorsCurve[curveIndex - 1]
                    if lastCurveEntry == nil then
                        priceFactor = curveEntry.priceFactor
                        yieldFactor = curveEntry.yieldFactor
                        decayFactor = curveEntry.decayFactor
                    else
                        local moistureAlpha = THUtils.getFactorInRange(lastCurveEntry.moisture, curveEntry.moisture, moisture)
                        priceFactor = THUtils.getValueInRange(lastCurveEntry.priceFactor, curveEntry.priceFactor, moistureAlpha)
                        yieldFactor = THUtils.getValueInRange(lastCurveEntry.yieldFactor, curveEntry.yieldFactor, moistureAlpha)
                        decayFactor = THUtils.getValueInRange(lastCurveEntry.decayFactor, curveEntry.decayFactor, moistureAlpha)
                    end
                    break
                end
            end
        end
    end
    return priceFactor, yieldFactor, decayFactor
end
THMoistureUtil = {}
function THMoistureUtil.calculateDewPoint(temp, humidity)
    local vA = 17.625
    local vB = 243.04
    local vH = math.log(humidity) + ((vA * temp) / (vB + temp))
    return math.min(temp, (vB * vH) / (vA - vH))
end
function THMoistureUtil.calculateHumidity(temp, dewPoint)
    local vA = 17.625
    local vB = 243.04
    local vD = math.exp((vA * dewPoint) / (vB + dewPoint))
    local vT = math.exp((vA * temp) / (vB + temp))
    return MathUtil.clamp(vD / vT, 0, 1)
end