-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THMoistureLoadingStation = {}
local THMoistureLoadingStation_mt = THUtils.createClass(THMoistureLoadingStation)
local debugFlagId = THDebugUtil.createDebugFlag("THLoadingStation")
THMoistureLoadingStation.debugFlagId = debugFlagId
local function getCustomData(self)
    return g_thMoistureSystem:getDataTable(self)
end
THMoistureLoadingStation.getCustomData = getCustomData
function THMoistureLoadingStation.newData(parent, owner, customMt)
    customMt = customMt or THMoistureLoadingStation_mt
    local self = setmetatable({}, customMt)
    self.owner = owner
    self.parent = parent
    self.moistureFillLevels = {}
    self.moistureFillLevelMapping = {}
    return self
end
function THMoistureLoadingStation:updateMoistureFillLevels()
    local owner = self.owner
    local station = self.parent
    local sourceStorages = station.sourceStorages
    local moistureFillLevels = THUtils.clearTable(self.moistureFillLevels)
    local moistureFillLevelMapping = THUtils.clearTable(self.moistureFillLevelMapping)
    if sourceStorages ~= nil then
        for _, storage in pairs(sourceStorages) do
            local storageData = THMoistureStorage.getCustomData(storage)
            local storageFillLevels = storage:getFillLevels()
            if storageData ~= nil and storageFillLevels ~= nil then
                for fillTypeIndex, fillLevel in pairs(storageFillLevels) do
                    if fillLevel > 0 and fillTypeIndex ~= FillType.UNKNOWN then
                        local fillTypeMoistureData = owner:getFillTypeMoistureData(fillTypeIndex)
                        local fillTypeMoisture = storageData:getFillTypeMoisture(fillTypeIndex)
                        local fillLevelIndex = moistureFillLevelMapping[fillTypeIndex]
                        if fillTypeMoistureData ~= nil and fillTypeMoisture >= 0 then
                            local fillTypeInfo = fillTypeMoistureData.fillType.info
                            local minMoisture = fillTypeMoistureData.minPercent
                            local maxMoisture = fillTypeMoistureData.maxPercent
                            local fillLevelData = nil
                            if fillLevelIndex == nil then
                                fillLevelData = {
                                    fillTypeIndex = fillTypeIndex,
                                    fillTypeTitle = fillTypeInfo.title,
                                    fillLevel = fillLevel,
                                    moisture = fillTypeMoisture
                                }
                                table.insert(moistureFillLevels, fillLevelData)
                                moistureFillLevelMapping[fillTypeIndex] = #moistureFillLevels
                            else
                                fillLevelData = moistureFillLevels[fillLevelIndex] -- should exist
                                local totalFillLevel = fillLevelData.fillLevel + fillLevel
                                local newFillPercent = MathUtil.clamp(fillLevel / totalFillLevel, 0, 1)
                                local oldFillPercent = 1 - newFillPercent
                                fillLevelData.fillLevel = totalFillLevel
                                fillLevelData.moisture = MathUtil.clamp((fillLevelData.moisture * oldFillPercent) + (fillTypeMoisture + newFillPercent), minMoisture, maxMoisture)
                            end
                        end
                    end
                end
            end
        end
    end
end
function THMoistureLoadingStation:getFillTypeMoisture(fillType)
    local owner = self.owner
    local fillTypeMoistureData = owner:getFillTypeMoistureData(fillType)
    if fillTypeMoistureData ~= nil then
        local fillLevelIndex = self.moistureFillLevelMapping[fillTypeMoistureData.fillType.index]
        if fillLevelIndex ~= nil then
            local moistureData = self.moistureFillLevels[fillLevelIndex]
            if moistureData ~= nil then
                return moistureData.moisture or -1
            end
        end
    end
    return -1
end
function THMoistureLoadingStation.hook_new(superFunc, ...)
    local function appendFunc(rStation, ...)
        if rStation ~= nil then
            THUtils.call(function()
                local customData = g_thMoistureSystem:createDataTable(rStation, false, THMoistureLoadingStation)
            end)
        end
        return rStation, ...
    end
    return appendFunc(superFunc(...))
end
function THMoistureLoadingStation:hook_addFillLevelToFillableObject(superFunc, object, fillUnitIndex, fillTypeIndex, deltaFill, ...)
    local customData = getCustomData(self)
    if customData ~= nil then
        THUtils.call(function()
            if object ~= nil and object.isa ~= nil
                and fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN
            then
                local fillTypeMoisture = customData:getFillTypeMoisture(fillTypeIndex)
                if fillTypeMoisture >= 0 then
                    if object:isa(Vehicle) and fillUnitIndex ~= nil
                        and object.thAA_setFillUnitInputMoisture ~= nil
                    then
                        object:thAA_setFillUnitInputMoisture(fillUnitIndex, fillTypeIndex, fillTypeMoisture)
                    elseif object:isa(UnloadTrigger) and object.target ~= nil then
                        local objectData = THMoistureUnloadingStation.getCustomData(object.target)
                        if objectData ~= nil and objectData.setFillTypeInputMoisture ~= nil then
                            objectData:setFillTypeInputMoisture(fillTypeIndex, fillTypeMoisture)
                        end
                    end
                end
            end
        end)
    end
    return superFunc(self, object, fillUnitIndex, fillTypeIndex, deltaFill, ...)
end
g_thMoistureSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("LoadingStation", "new", THMoistureLoadingStation, nil, false)
    THUtils.setFunctionHook("LoadingStation", "addFillLevelToFillableObject", THMoistureLoadingStation)
end)