-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thMoistureSystem.modName
THMoistureStorage = {}
local THMoistureStorage_mt = THUtils.createClass(THMoistureStorage)
local debugFlagId = THDebugUtil.createDebugFlag("THStorage")
THMoistureStorage.debugFlagId = debugFlagId
local function getCustomData(self)
    return g_thMoistureSystem:getDataTable(self)
end
THMoistureStorage.getCustomData = getCustomData
function THMoistureStorage.newData(parent, owner, customMt)
    customMt = customMt or THMoistureStorage_mt
    local self = setmetatable({}, customMt)
    self.isServer = parent.isServer
    self.isClient = parent.isClient
    self.owner = owner
    self.parent = parent
    self.fillTypeMoisture = {}
    self.lastSentFillTypeMoisture = {}
    self.moistureChangedListeners = {}
    self.inputMoistureData = {}
    self.dirtyFlag = parent:getNextDirtyFlag()
    local function onFillLevelChanged(pFillTypeIndex, pAppliedDelta, ...)
        return self:onFillLevelChanged(pFillTypeIndex, pAppliedDelta, ...)
    end
    parent:addFillLevelChangedListeners(onFillLevelChanged)
    return self
end
function THMoistureStorage:onFillLevelChanged(fillTypeIndex, appliedDelta, ...)
    self:updateFillTypeMoisture(fillTypeIndex, nil, appliedDelta)
end
function THMoistureStorage:getFillTypeMoisture(fillType)
    local owner = self.owner
    local storage = self.parent
    local fillTypeMoistureData = owner:getFillTypeMoistureData(fillType)
    local storageMoisture = -1
    local storageFillLevel = 0
    if fillTypeMoistureData ~= nil then
        local fillTypeIndex = fillTypeMoistureData.fillType.index
        local minMoisture = fillTypeMoistureData.minPercent
        local maxMoisture = fillTypeMoistureData.maxPercent
        local defMoisture = fillTypeMoistureData.defPercent
        local curMoisture = THUtils.getNoNil(self.fillTypeMoisture[fillTypeIndex], -1)
        storageFillLevel = storage:getFillLevel(fillTypeIndex) or 0
        if storageFillLevel > 0 then
            if curMoisture < 0 then
                storageMoisture = defMoisture
            else
                storageMoisture = MathUtil.clamp(curMoisture, minMoisture, maxMoisture)
            end
        end
    end
    return storageMoisture, storageFillLevel
end
function THMoistureStorage:setFillTypeMoisture(fillType, moisture, noEventSend)
    if THUtils.argIsValid(type(moisture) == "number", "moisture", moisture) then
        local owner = self.owner
        local storage = self.parent
        local fillTypeMoistureData = owner:getFillTypeMoistureData(fillType)
        if fillTypeMoistureData ~= nil then
            local fillTypeIndex = fillTypeMoistureData.fillType.index
            local fillTypeName = fillTypeMoistureData.fillType.name
            local storageFillLevel = storage:getFillLevel(fillTypeIndex) or 0
            local oldMoisture = self.fillTypeMoisture[fillTypeIndex] or -1
            local newMoisture = oldMoisture
            if storageFillLevel > 0 then
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local defMoisture = fillTypeMoistureData.defPercent
                if moisture < 0 then
                    newMoisture = defMoisture
                else
                    newMoisture = MathUtil.clamp(moisture, minMoisture, maxMoisture)
                end
            else
                newMoisture = -1
            end
            if math.abs(newMoisture - oldMoisture) > 0.000001 then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THDebugUtil.printMsg(debugFlagId, nil, "Moisture values for fillType: %s", fillTypeName)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Old moisture: %s", oldMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "- New moisture: %s", newMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                end
                self.fillTypeMoisture[fillTypeIndex] = newMoisture
                for _, onChangeFunc in pairs(self.moistureChangedListeners) do
                    THUtils.call(onChangeFunc, storage, newMoisture, oldMoisture)
                end
            end
            if self.isServer and noEventSend ~= true then
                local currentMoisture = self.fillTypeMoisture[fillTypeIndex] or -1
                local lastSentMoisture = self.lastSentFillTypeMoisture[fillTypeIndex]
                if lastSentMoisture == nil or math.abs(currentMoisture - lastSentMoisture) > 0.0001 then
                    storage:raiseDirtyFlags(self.dirtyFlag)
                end
            end
        end
    end
end
function THMoistureStorage:updateFillTypeMoisture(fillType, inputMoisture, inputDelta, noEventSend)
    if self.isServer then
        local owner = self.owner
        local storage = self.parent
        local fillTypeMoistureData = owner:getFillTypeMoistureData(fillType)
        if fillTypeMoistureData ~= nil then
            local fillTypeIndex = fillTypeMoistureData.fillType.index
            local fillTypeName = fillTypeMoistureData.fillType.name
            local storageFillLevel = storage:getFillLevel(fillTypeIndex) or 0
            local oldMoisture = THUtils.getNoNil(self.fillTypeMoisture[fillTypeIndex], -1)
            local newMoisture = oldMoisture
            if storageFillLevel <= 0 then
                newMoisture = -1
            else
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local defMoisture = fillTypeMoistureData.defPercent
                if inputDelta ~= nil then
                    inputDelta = MathUtil.clamp(inputDelta, -storageFillLevel, storageFillLevel)
                end
                if inputMoisture == nil then
                    local inputMoistureData = self.inputMoistureData[fillTypeIndex]
                    if inputMoistureData ~= nil and inputMoistureData.moisture ~= nil
                        and inputMoistureData.moisture >= 0
                    then
                        inputMoisture = inputMoistureData.moisture
                    end
                end
                if inputMoisture ~= nil and inputDelta ~= nil then
                    if inputMoisture < 0 then
                        inputMoisture = defMoisture
                    else
                        inputMoisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                    end
                    if oldMoisture < 0 then
                        newMoisture = inputMoisture
                    elseif inputDelta > 0 then
                        local newFillPercent = MathUtil.clamp(inputDelta / storageFillLevel, 0, 1)
                        local oldFillPercent = 1 - newFillPercent
                        newMoisture = MathUtil.clamp((oldMoisture * oldFillPercent) + (inputMoisture * newFillPercent), minMoisture, maxMoisture)
                    else
                        newMoisture = MathUtil.clamp(oldMoisture, minMoisture, maxMoisture)
                    end
                else
                    if oldMoisture < 0 then
                        newMoisture = defMoisture
                    elseif inputDelta ~= nil and inputDelta > 0 then
                        local newFillPercent = MathUtil.clamp(inputDelta / storageFillLevel, 0, 1)
                        local oldFillPercent = 1 - newFillPercent
                        newMoisture = MathUtil.clamp((oldMoisture * oldFillPercent) + (defMoisture * newFillPercent), minMoisture, maxMoisture)
                    else
                        newMoisture = MathUtil.clamp(oldMoisture, minMoisture, maxMoisture)
                    end
                end
            end
            if math.abs(newMoisture - oldMoisture) > 0.000001 then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THDebugUtil.printMsg(debugFlagId, nil, "Moisture values for fillType: %s", fillTypeName)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Old moisture: %s", oldMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "- New moisture: %s", newMoisture)
                    if inputMoisture ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input moisture: %s", inputMoisture)
                    end
                    if inputDelta ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input delta: %s", inputDelta)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                end
                self.fillTypeMoisture[fillTypeIndex] = newMoisture
                for _, onChangeFunc in pairs(self.moistureChangedListeners) do
                    THUtils.call(onChangeFunc, storage, newMoisture, oldMoisture)
                end
            end
            if noEventSend ~= true then
                local currentMoisture = self.fillTypeMoisture[fillTypeIndex] or -1
                local lastSentMoisture = self.lastSentFillTypeMoisture[fillTypeIndex]
                if lastSentMoisture == nil
                    or math.abs(currentMoisture - lastSentMoisture) > 0.0001
                then
                    storage:raiseDirtyFlags(self.dirtyFlag)
                end
            end
        end
    end
end
function THMoistureStorage:setFillTypeInputMoisture(fillType, inputMoisture)
    if THUtils.argIsValid(type(inputMoisture) == "number", "inputMoisture", inputMoisture) then
        local owner = self.owner
        local storage = self.parent
        local thFillTypeManager = owner.thFillTypeManager
        local fillTypeInfo = thFillTypeManager:getFillType(fillType)
        if fillTypeInfo ~= nil and fillTypeInfo.index ~= nil then
            local fillTypeMoistureData = owner:getFillTypeMoistureData(fillTypeInfo.index)
            if fillTypeMoistureData == nil then
                self.inputMoistureData[fillTypeInfo.index] = nil
            else
                local inputMoistureData = THUtils.createSubTable(self.inputMoistureData, fillTypeInfo.index, true)
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                if inputMoisture >= 0 then
                    inputMoistureData.fillTypeIndex = fillTypeInfo.index
                    inputMoistureData.moisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                    inputMoistureData.resetTime = 50
                    THDebugUtil.printMsg(debugFlagId, nil, "Input moisture set for fillType: %s", fillTypeInfo.name)
                    THDebugUtil.printTable(debugFlagId, nil, inputMoistureData, 1)
                    storage:raiseActive()
                end
            end
        end
    end
end
function THMoistureStorage:readStorageFromStream(streamId, connection)
    local storage = self.parent
    local sortedFillTypes = storage.sortedFillTypes
    local numFillTypes = 0
    if sortedFillTypes ~= nil then
        numFillTypes = #sortedFillTypes
    end
    if numFillTypes > 0 then
        for i = 1, numFillTypes do
            local fillTypeIndex = sortedFillTypes[i]
            self.fillTypeMoisture[fillTypeIndex] = streamReadFloat32(streamId)
        end
    end
end
function THMoistureStorage:writeStorageToStream(streamId, connection)
    local storage = self.parent
    local sortedFillTypes = storage.sortedFillTypes
    local numFillTypes = 0
    if sortedFillTypes ~= nil then
        numFillTypes = #sortedFillTypes
    end
    if numFillTypes > 0 then
        for i = 1, numFillTypes do
            local fillTypeIndex = sortedFillTypes[i]
            local fillTypeMoisture = THUtils.getNoNil(self.fillTypeMoisture[fillTypeIndex], -1)
            streamWriteFloat32(streamId, fillTypeMoisture)
            self.lastSentFillTypeMoisture[fillTypeIndex] = fillTypeMoisture
        end
    end
end
function THMoistureStorage.hook_registerSavegameXMLPaths(superFunc, xmlSchema, xmlPath, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local xmlBasePath = string.format("%s.%s", xmlPath, thModName)
            THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlBasePath .. ".fillType(?)", "#name", "Name of moisture fillType")
            THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlBasePath .. ".fillType(?)", "#moisture", "Current fillType moisture percentage")
        end)
        return ...
    end
    return appendFunc(superFunc(xmlSchema, xmlPath))
end
function THMoistureStorage.hook_new(superFunc, ...)
    local function appendFunc(rStorage, ...)
        if rStorage ~= nil then
            THUtils.call(function()
                local customData = g_thMoistureSystem:createDataTable(rStorage, false, THMoistureStorage)
            end)
        end
        return rStorage, ...
    end
    return appendFunc(superFunc(...))
end
function THMoistureStorage:hook_update(superFunc, dt, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                local hasInputMoisture = false
                for _, inputMoistureData in pairs(customData.inputMoistureData) do
                    if inputMoistureData.resetTime ~= nil and inputMoistureData.resetTime > 0 then
                        inputMoistureData.resetTime = inputMoistureData.resetTime - dt
                        if inputMoistureData.resetTime <= 0 then
                            THUtils.clearTable(inputMoistureData)
                        else
                            hasInputMoisture = true
                        end
                    elseif next(inputMoistureData) ~= nil then
                        THUtils.clearTable(inputMoistureData)
                    end
                end
                if hasInputMoisture then
                    self:raiseActive()
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THMoistureStorage:hook_readStream(superFunc, streamId, connection, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and connection:getIsServer() then
            customData:readStorageFromStream(streamId, connection)
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, ...))
end
function THMoistureStorage:hook_writeStream(superFunc, streamId, connection, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and not connection:getIsServer() then
            customData:writeStorageToStream(streamId, connection)
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, ...))
end
function THMoistureStorage:hook_readUpdateStream(superFunc, streamId, timeStamp, connection, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and connection:getIsServer() then
            if streamReadBool(streamId) then -- dirtyMask
                customData:readStorageFromStream(streamId, connection)
            end
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, timeStamp, connection, ...))
end
function THMoistureStorage:hook_writeUpdateStream(superFunc, streamId, connection, dirtyMask, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and not connection:getIsServer() then
            if streamWriteBool(streamId, bitAND(dirtyMask, customData.dirtyFlag) ~= 0) then
                customData:writeStorageToStream(streamId, connection)
            end
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, dirtyMask, ...))
end
function THMoistureStorage:hook_loadFromXMLFile(superFunc, xmlFile, xmlKey, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        THUtils.call(function()
            if self.isServer and customData ~= nil then
                local owner = customData.owner
                local thFillTypeManager = owner.thFillTypeManager
                local xmlBaseKey = string.format("%s.%s", xmlKey, thModName)
                xmlFile:iterate(xmlBaseKey .. ".fillType", function(_, pFillTypeKey)
                    local fillTypeName = THUtils.getXMLValue(xmlFile, pFillTypeKey, "#name")
                    local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
                    if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
                        local fillTypeIndex = fillTypeInfo.index
                        local fillTypeMoisture = THUtils.getXMLValue(xmlFile, pFillTypeKey, "#moisture")
                        if fillTypeMoisture ~= nil then
                            customData:setFillTypeMoisture(fillTypeIndex, fillTypeMoisture, true)
                        end
                    end
                end)
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, xmlFile, xmlKey, ...))
end
function THMoistureStorage:hook_saveToXMLFile(superFunc, xmlFile, xmlKey, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if self.isServer and customData ~= nil then
            THUtils.call(function()
                local owner = customData.owner
                local thFillTypeManager = owner.thFillTypeManager
                local xmlBaseKey = string.format("%s.%s", xmlKey, thModName)
                local fillTypeXMLIndex = 0
                for fillTypeIndex, fillTypeMoisture in pairs(customData.fillTypeMoisture) do
                    local fillTypeInfo = thFillTypeManager:getFillType(fillTypeIndex)
                    local storageFillLevel = self:getFillLevel(fillTypeIndex) or 0
                    if storageFillLevel > 0 and fillTypeMoisture >= 0
                        and fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN
                    then
                        local fillTypeKey = string.format("%s.fillType(%d)", xmlBaseKey, fillTypeXMLIndex)
                        xmlFile:setString(fillTypeKey .. "#name", fillTypeInfo.name)
                        xmlFile:setFloat(fillTypeKey .. "#moisture", fillTypeMoisture)
                        fillTypeXMLIndex = fillTypeXMLIndex + 1
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, xmlFile, xmlKey, ...))
end
function THMoistureStorage:hook_addUnloadingStation(superFunc, unloadingStation, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                if unloadingStation ~= nil and self.unloadingStations[unloadingStation] ~= nil then
                    local placeable = unloadingStation.owningPlaceable
                    if placeable ~= nil and placeable.thAA_addMoistureStorage ~= nil then
                        placeable:thAA_addMoistureStorage(self)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, unloadingStation, ...))
end
function THMoistureStorage:hook_removeUnloadingStation(superFunc, unloadingStation, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                if unloadingStation ~= nil and self.unloadingStations[unloadingStation] == nil then
                    local placeable = unloadingStation.owningPlaceable
                    if placeable ~= nil and placeable.thAA_removeMoistureStorage ~= nil then
                        placeable:thAA_removeMoistureStorage(self)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, unloadingStation, ...))
end
g_thMoistureSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("Storage", "registerSavegameXMLPaths", THMoistureStorage, nil, false)
    THUtils.setFunctionHook("Storage", "new", THMoistureStorage, nil, false)
    THUtils.setFunctionHook("Storage", "update", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "readStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "writeStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "readUpdateStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "writeUpdateStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "loadFromXMLFile", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "saveToXMLFile", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "addUnloadingStation", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "removeUnloadingStation", THMoistureStorage)
end)