-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THMoistureUnloadingStation = {}
local THMoistureUnloadingStation_mt = THUtils.createClass(THMoistureUnloadingStation, nil, false) -- unprotected/changeable
local debugFlagId = THDebugUtil.createDebugFlag("THUnloadingStation")
THMoistureUnloadingStation.debugFlagId = debugFlagId
local function getCustomData(self)
    return g_thMoistureSystem:getDataTable(self)
end
THMoistureUnloadingStation.getCustomData = getCustomData
THMoistureSellingStation = {}
function THMoistureUnloadingStation.newData(parent, owner, customMt)
    customMt = customMt or THMoistureUnloadingStation_mt
    local self = setmetatable({}, customMt)
    self.owner = owner
    self.parent = parent
    self.inputMoistureData = {}
    self.hasMoistureMoneyChange = false
    return self
end
function THMoistureUnloadingStation:setFillTypeInputMoisture(fillType, inputMoisture)
    if THUtils.argIsValid(type(inputMoisture) == "number", "inputMoisture", inputMoisture) then
        local owner = self.owner
        local station = self.parent
        local thFillTypeManager = owner.thFillTypeManager
        local fillTypeInfo = thFillTypeManager:getFillType(fillType)
        if fillTypeInfo ~= nil and fillTypeInfo.index ~= nil then
            local fillTypeIndex = fillTypeInfo.index
            local fillTypeMoistureData = owner:getFillTypeMoistureData(fillTypeIndex)
            if fillTypeMoistureData == nil then
                self.inputMoistureData[fillTypeIndex] = nil
            else
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local inputMoistureData = THUtils.createSubTable(self.inputMoistureData, fillTypeIndex, true)
                if inputMoisture >= 0 then
                    inputMoistureData.fillTypeIndex = fillTypeIndex
                    inputMoistureData.moisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                    inputMoistureData.resetTime = 50
                    THDebugUtil.printMsg(debugFlagId, nil, "Input moisture set for fillType: %s", fillTypeInfo.name)
                    THDebugUtil.printTable(debugFlagId, nil, inputMoistureData, 1)
                    local targetStorages = station.targetStorages
                    if targetStorages ~= nil then
                        for _, storage in pairs(targetStorages) do
                            local storageData = THMoistureStorage.getCustomData(storage)
                            if storageData ~= nil then
                                storageData:setFillTypeInputMoisture(fillType, inputMoisture)
                            end
                        end
                    end
                    station:raiseActive()
                end
            end
        end
    end
end
function THMoistureUnloadingStation:getFillTypeInputMoisture(fillType)
    local owner = self.owner
    local fillTypeMoistureData = owner:getFillTypeMoistureData(fillType)
    if fillTypeMoistureData ~= nil then
        local fillTypeInfo = fillTypeMoistureData.fillType.info
        local inputMoistureData = self.inputMoistureData[fillTypeInfo.index]
        if inputMoistureData ~= nil then
            return inputMoistureData.moisture or -1
        end
    end
    return -1
end
function THMoistureUnloadingStation.hook_new(superFunc, ...)
    local function appendFunc(rStation, ...)
        if rStation ~= nil then
            THUtils.call(function()
                local customData = g_thMoistureSystem:createDataTable(rStation, false, THMoistureUnloadingStation)
            end)
        end
        return rStation, ...
    end
    return appendFunc(superFunc(...))
end
function THMoistureSellingStation:hook_update(superFunc, dt, ...)
    local customData = getCustomData(self)
    local doMoneyChange = false
    if customData ~= nil and customData.hasMoistureMoneyChange
        and self.lastMoneyChange ~= nil and self.lastMoneyChange > 0
        and self.lastMoneyChangeFarmId ~= nil
    then
        doMoneyChange = true
    end
    local function appendFunc(...)
        if customData ~= nil and doMoneyChange then
            THUtils.call(function()
                local owner = customData.owner
                local mission = owner.mission
                if self.lastMoneyChange <= 0 then
                    local moneyChangeType = owner.moneyChangeTypes.harvestMoisture
                    mission:showMoneyChange(moneyChangeType, nil, false, self.lastMoneyChangeFarmId)
                    customData.hasMoistureMoneyChange = false
                end
                local hasInputMoisture = false
                for _, inputMoistureData in pairs(customData.inputMoistureData) do
                    if inputMoistureData.resetTime ~= nil and inputMoistureData.resetTime > 0 then
                        inputMoistureData.resetTime = inputMoistureData.resetTime - dt
                        if inputMoistureData.resetTime <= 0 then
                            THUtils.clearTable(inputMoistureData)
                        else
                            hasInputMoisture = true
                        end
                    elseif next(inputMoistureData) ~= nil then
                        THUtils.clearTable(inputMoistureData)
                    end
                end
                if hasInputMoisture then
                    self:raiseActive()
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THMoistureSellingStation:hook_sellFillType(superFunc, farmId, deltaFill, fillTypeIndex, toolType, extraAttributes, ...)
    local customData = getCustomData(self)
    local function appendFunc(rPrice, ...)
        if customData ~= nil and rPrice ~= nil and rPrice > 0
            and fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN
        then
            THUtils.call(function()
                local owner = customData.owner
                local mission = owner.mission
                local thFillTypeManager = owner.thFillTypeManager
                local fillTypeMoisture = customData:getFillTypeInputMoisture(fillTypeIndex)
                local moneyChangeType = owner.moneyChangeTypes.harvestMoisture
                if fillTypeMoisture >= 0 then
                    local priceFactor = owner:getFillTypeMoistureFactors(fillTypeIndex, fillTypeMoisture)
                    local moistureBonus = (rPrice * priceFactor) - rPrice
                    if math.abs(moistureBonus) > 0.001 then
                        mission:addMoney(moistureBonus, farmId, moneyChangeType, true)
                        customData.hasMoistureMoneyChange = true
                        if THDebugUtil.getIsEnabled(debugFlagId) then
                            local fillTypeName = thFillTypeManager:getFillTypeName(fillTypeIndex)
                            THDebugUtil.printMsg(debugFlagId, nil, "Moisture price for fillType: %s", fillTypeName)
                            THDebugUtil.printMsg(debugFlagId, nil, "- Moisture percentage: %s", fillTypeMoisture)
                            THDebugUtil.printMsg(debugFlagId, nil, "- Price factor: %s", priceFactor)
                            THDebugUtil.printMsg(debugFlagId, nil, "- Sell price: %s", rPrice)
                            THDebugUtil.printMsg(debugFlagId, nil, "- Moisture bonus: -%s", moistureBonus)
                            THDebugUtil.printMsg(debugFlagId, nil, "")
                        end
                    end
                end
            end)
        end
        return rPrice, ...
    end
    return appendFunc(superFunc(self, farmId, deltaFill, fillTypeIndex, toolType, extraAttributes, ...))
end
g_thMoistureSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("UnloadingStation", "new", THMoistureUnloadingStation, nil, false)
    THUtils.setFunctionHook("SellingStation", "update", THMoistureSellingStation)
    THUtils.setFunctionHook("SellingStation", "sellFillType", THMoistureSellingStation)
end)