-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thI18n = g_thMoistureSystem.i18n
local thModName = g_thMoistureSystem.modName
THPlaceableMoistureStorage = {}
local debugFlagId = THDebugUtil.createDebugFlag("THPlaceableMoistureStorage")
THPlaceableMoistureStorage.debugFlagId = debugFlagId
THPlaceableMoistureStorage.MOD_NAME = thModName
THPlaceableMoistureStorage.SPEC_NAME = "moistureStorage"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, thModName, THPlaceableMoistureStorage.SPEC_NAME, verbose)
end
THPlaceableMoistureStorage.getSpecTable = getSpecTable
function THPlaceableMoistureStorage.initSpecialization()
end
function THPlaceableMoistureStorage.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THPlaceable, specializations)
        and SpecializationUtil.hasSpecialization(PlaceableInfoTrigger, specializations)
end
function THPlaceableMoistureStorage.registerXMLPaths(xmlSchema, xmlPath)
    xmlSchema:setXMLSpecializationType("THMoistureStorage")
    local basePath = THPlaceable.getConfigSchemaPath(THPlaceableMoistureStorage.SPEC_NAME)
    THStorageDryer.registerXMLPaths(xmlSchema, basePath .. ".storageDryer")
    xmlSchema:setXMLSpecializationType()
end
function THPlaceableMoistureStorage.registerSavegameXMLPaths(xmlSchema, xmlPath)
    xmlSchema:setXMLSpecializationType("THMoistureStorage")
    xmlSchema:setXMLSpecializationType()
end
function THPlaceableMoistureStorage.registerEvents(placeableType)
end
function THPlaceableMoistureStorage.registerEventListeners(placeableType)
    SpecializationUtil.registerEventListener(placeableType, "onPreLoad", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onLoad", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onDelete", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onReadStream", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onWriteStream", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onMinuteChanged", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onDayChanged", THPlaceableMoistureStorage)
end
function THPlaceableMoistureStorage.registerFunctions(placeableType)
    SpecializationUtil.registerFunction(placeableType, "thAA_addMoistureStorage", THPlaceableMoistureStorage.thAA_addMoistureStorage)
    SpecializationUtil.registerFunction(placeableType, "thAA_removeMoistureStorage", THPlaceableMoistureStorage.thAA_removeMoistureStorage)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateMoistureStorageFillLevels", THPlaceableMoistureStorage.thAA_updateMoistureStorageFillLevels)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateMoistureHudBoxValues", THPlaceableMoistureStorage.thAA_updateMoistureHudBoxValues)
    SpecializationUtil.registerFunction(placeableType, "thAA_addMoistureHudBoxEntries", THPlaceableMoistureStorage.thAA_addMoistureHudBoxEntries)
end
function THPlaceableMoistureStorage.registerOverwrittenFunctions(placeableType)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "setOwnerFarmId", THPlaceableMoistureStorage.setOwnerFarmId)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "getNeedMinuteChanged", THPlaceableMoistureStorage.getNeedMinuteChanged)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "getNeedDayChanged", THPlaceableMoistureStorage.getNeedDayChanged)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "updateInfo", THPlaceableMoistureStorage.updateInfo)
end
function THPlaceableMoistureStorage:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.moistureStorages = {}
    specTable.moistureFillLevels = {}
    specTable.moistureFillLevelMapping = {}
    if specTable.moistureHudBox == nil then
        specTable.moistureHudBox = g_thMoistureSystem.hud.infoDisplay:createBox(KeyValueInfoHUDBox)
    end
    specTable.moistureHudBoxValues = {}
    specTable.isMoistureHudBoxEnabled = true
    specTable.hudTitles = {
        moisture = thI18n:getText("thUI_hudTitleMoisture"),
        storage = thI18n:getText("thUI_storage"),
        dryerStatus = thI18n:getText("thStorageDryerStatus")
    }
end
function THPlaceableMoistureStorage:onLoad(savegame)
    local specTable = getSpecTable(self)
    local siloSpec = self.spec_silo
    local infoTriggerSpec = self.spec_infoTrigger
    local components = self.components
    local xmlFile = self.xmlFile
    local xmlBaseKey = THPlaceable.getConfigBaseKey(THPlaceableMoistureStorage.SPEC_NAME)
    local customEnv = self.customEnvironment
    local i3dMappings = self.i3dMappings
    local storageDryerXMLKey = xmlBaseKey .. ".storageDryer"
    local dryerTriggerNode = 0
    if siloSpec ~= nil and infoTriggerSpec ~= nil
        and infoTriggerSpec.infoTrigger ~= nil
        and self.getFillLevels ~= nil
    then
        local fillLevelList = self:getFillLevels()
        local hasMoistureFillType = false
        if fillLevelList ~= nil then
            for fillTypeIndex in pairs(fillLevelList) do
                if g_thMoistureSystem:getFillTypeMoistureData(fillTypeIndex) ~= nil then
                    hasMoistureFillType = true
                    break
                end
            end
        end
        if hasMoistureFillType and infoTriggerSpec.infoTrigger > 0 then
            dryerTriggerNode = infoTriggerSpec.infoTrigger
        end
    end
    if xmlFile:hasProperty(storageDryerXMLKey) or dryerTriggerNode > 0 then
        local storageDryer = THStorageDryer.new(self)
        if storageDryer ~= nil and storageDryer:getIsEnabled() then
            if storageDryer:load(components, xmlFile, storageDryerXMLKey, customEnv, i3dMappings, dryerTriggerNode) then
                specTable.storageDryer = storageDryer
            else
                THUtils.stackMsg("Storage dryer failed to load")
                storageDryer:delete()
            end
        end
    end
end
function THPlaceableMoistureStorage:onDelete()
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        specTable.storageDryer:delete()
        specTable.storageDryer = nil
    end
end
function THPlaceableMoistureStorage:onFinalizePlacement()
    local specTable = getSpecTable(self)
    local ownerFarmId = self:getOwnerFarmId()
    local storageDryer = specTable.storageDryer
    if storageDryer ~= nil then
        storageDryer:register(true)
        storageDryer:setOwnerFarmId(ownerFarmId)
        storageDryer:setDryerStatus(THStorageDryer.DRYER_STATUS.STOPPED, true)
    end
end
function THPlaceableMoistureStorage:onReadStream(streamId, connection, ...)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        local storageDryerObjectId = NetworkUtil.readNodeObjectId(streamId)
        specTable.storageDryer:readStream(streamId, connection, ...)
        g_client:finishRegisterObject(specTable.storageDryer, storageDryerObjectId)
    end
end
function THPlaceableMoistureStorage:onWriteStream(streamId, connection, ...)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        local storageDryerObjectId = NetworkUtil.getObjectId(specTable.storageDryer)
        NetworkUtil.writeNodeObjectId(streamId, storageDryerObjectId)
        specTable.storageDryer:writeStream(streamId, connection, ...)
        g_server:registerObjectInStream(connection, specTable.storageDryer)
    end
end
function THPlaceableMoistureStorage:loadFromXMLFile(xmlFile, xmlKey)
end
function THPlaceableMoistureStorage:saveToXMLFile(xmlFile, xmlKey, usedModNames)
end
function THPlaceableMoistureStorage:onMinuteChanged(currentMinute)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        specTable.storageDryer:updateStorageDryer(1, specTable.moistureStorages)
    end
end
function THPlaceableMoistureStorage:onDayChanged(currentDay)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        specTable.storageDryer:processDryingCosts()
    end
end
function THPlaceableMoistureStorage:thAA_addMoistureStorage(storage)
    local specTable = getSpecTable(self)
    if storage ~= nil and storage.isa and storage:isa(Storage) then
        specTable.moistureStorages[storage] = storage
    end
end
function THPlaceableMoistureStorage:thAA_removeMoistureStorage(storage)
    local specTable = getSpecTable(self)
    if storage ~= nil then
        specTable.moistureStorages[storage] = nil
    end
end
function THPlaceableMoistureStorage:thAA_updateMoistureStorageFillLevels()
    local specTable = getSpecTable(self)
    local moistureFillLevels = THUtils.clearTable(specTable.moistureFillLevels)
    local moistureFillLevelMapping = THUtils.clearTable(specTable.moistureFillLevelMapping)
    for _, storage in pairs(specTable.moistureStorages) do
        local storageData = THMoistureStorage.getCustomData(storage)
        local storageFillLevels = storage:getFillLevels()
        if storageFillLevels ~= nil and storageData ~= nil then
            for fillTypeIndex, fillLevel in pairs(storageFillLevels) do
                if fillLevel > 0 and fillTypeIndex ~= FillType.UNKNOWN then
                    local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillTypeIndex)
                    local fillTypeMoisture = storageData:getFillTypeMoisture(fillTypeIndex)
                    if fillTypeMoistureData ~= nil and fillTypeMoisture >= 0 then
                        local fillTypeInfo = fillTypeMoistureData.fillType.info
                        local fillLevelIndex = moistureFillLevelMapping[fillTypeIndex]
                        local fillLevelData = nil
                        if fillLevelIndex == nil then
                            fillLevelData = {
                                fillTypeIndex = fillTypeIndex,
                                fillTypeTitle = fillTypeInfo.title,
                                fillLevel = fillLevel,
                                moisture = fillTypeMoisture
                            }
                            table.insert(moistureFillLevels, fillLevelData)
                            moistureFillLevelMapping[fillTypeIndex] = #moistureFillLevels
                        else
                            fillLevelData = moistureFillLevels[fillLevelIndex] -- should exist
                            local totalFillLevel = fillLevelData.fillLevel + fillLevel
                            local minMoisture = fillTypeMoistureData.minPercent
                            local maxMoisture = fillTypeMoistureData.maxPercent
                            local newFillPercent = MathUtil.clamp(fillLevel / totalFillLevel, 0, 1)
                            local oldFillPercent = 1 - newFillPercent
                            fillLevelData.fillLevel = totalFillLevel
                            fillLevelData.moisture = MathUtil.clamp((fillLevelData.moisture * oldFillPercent) + (fillTypeMoisture * newFillPercent), minMoisture, maxMoisture)
                        end
                    end
                end
            end
        end
    end
    if #moistureFillLevels > 0 then
        table.sort(moistureFillLevels, function(arg1, arg2)
            return arg1.fillLevel > arg2.fillLevel
        end)
    end
    return moistureFillLevels
end
function THPlaceableMoistureStorage:thAA_updateMoistureHudBoxValues()
    local specTable = getSpecTable(self)
    local hudBoxValues = THUtils.clearTable(specTable.moistureHudBoxValues)
    local moistureFillLevels = self:thAA_updateMoistureStorageFillLevels()
    local numFillLevelEntries = #moistureFillLevels
    if numFillLevelEntries > 0 then
        for fillLevelIndex = 1, numFillLevelEntries do
            local fillLevelData = moistureFillLevels[fillLevelIndex]
            local moistureText = THUtils.formatPercent(fillLevelData.moisture, 1)
            local hudBoxData = {
                fillType = fillLevelData.fillTypeIndex,
                title = fillLevelData.fillTypeTitle,
                text = moistureText,
                accentuate = false
            }
            table.insert(hudBoxValues, hudBoxData)
        end
    end
    return hudBoxValues
end
function THPlaceableMoistureStorage:thAA_addMoistureHudBoxEntries(hudBox)
    local specTable = getSpecTable(self)
    local isMoistureHudBox = hudBox == specTable.moistureHudBox
    if isMoistureHudBox then
        hudBox:clear()
    end
    local hudBoxValues = self:thAA_updateMoistureHudBoxValues()
    local numHudBoxValues = #hudBoxValues
    if numHudBoxValues > 0 then
        local hudBoxTitle = specTable.hudTitles.moisture
        local siloName = self:getName() or specTable.hudTitles.storage
        if isMoistureHudBox then
            hudBox:setTitle(hudBoxTitle)
            hudBox:addLine(siloName, "", true)
        end
        for valueIndex = 1, numHudBoxValues do
            local hudBoxData = hudBoxValues[valueIndex]
            hudBox:addLine("- " .. hudBoxData.title, hudBoxData.text, hudBoxData.accentuate, hudBoxData.accentuateColor)
            hudBoxValues[valueIndex] = nil
        end
        if specTable.storageDryer ~= nil then
            local dryerStatus = specTable.storageDryer:getCurrentDryerStatus()
            if dryerStatus ~= THStorageDryer.DRYER_STATUS.DISABLED then
                local dryerStatusText = specTable.storageDryer:getDryerStatusText()
                if dryerStatusText ~= nil then
                    local statusTitle = specTable.hudTitles.dryerStatus
                    hudBox:addLine(statusTitle, dryerStatusText)
                end
            end
        end
        if isMoistureHudBox then
            hudBox:showNextFrame()
        end
    end
end
function THPlaceableMoistureStorage:setOwnerFarmId(superFunc, farmId, noEventSend, ...)
    local specTable = getSpecTable(self)
    local function appendFunc(...)
        if specTable.storageDryer ~= nil then
            THUtils.call(function()
                specTable.storageDryer:setOwnerFarmId(farmId, noEventSend)
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, farmId, noEventSend, ...))
end
function THPlaceableMoistureStorage:getNeedMinuteChanged(superFunc, ...)
    local function appendFunc(rNeedChange, ...)
        local specTable = getSpecTable(self)
        if specTable.storageDryer ~= nil then
            rNeedChange = true
        end
        return rNeedChange, ...
    end
    return appendFunc(superFunc(self, ...))
end
function THPlaceableMoistureStorage:getNeedDayChanged(superFunc, ...)
    local function appendFunc(rNeedChange, ...)
        local specTable = getSpecTable(self)
        if specTable.storageDryer ~= nil then
            rNeedChange = true
        end
        return rNeedChange, ...
    end
    return appendFunc(superFunc(self, ...))
end
function THPlaceableMoistureStorage:updateInfo(superFunc, info, ...)
    local specTable = getSpecTable(self)
    local function appendFunc(...)
        if specTable.isMoistureHudBoxEnabled then
            THUtils.call(function()
                self:thAA_addMoistureHudBoxEntries(specTable.moistureHudBox)
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, info, ...))
end