-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thMoistureSystem.modName
local thFillTypeManager = g_thMoistureSystem.thFillTypeManager
THMoistureFillUnit = {}
local debugFlagId = THDebugUtil.createDebugFlag("THFillUnit")
THMoistureFillUnit.debugFlagId = debugFlagId
THMoistureFillUnit.MOD_NAME = thModName
THMoistureFillUnit.SPEC_NAME = "moistureFillUnit"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, thModName, THMoistureFillUnit.SPEC_NAME, verbose)
end
THMoistureFillUnit.getSpecTable = getSpecTable
function THMoistureFillUnit.initSpecialization()
    local schemaSavegame = Vehicle.xmlSchemaSavegame
    local savegameBasePath = THVehicle.getSavegameSchemaPath(THMoistureFillUnit.SPEC_NAME)
    local fillUnitPath = savegameBasePath .. ".fillUnit(?)"
    THUtils.registerXMLPath(schemaSavegame, XMLValueType.INT, fillUnitPath, "#index", "Fill unit index")
    THUtils.registerXMLPath(schemaSavegame, XMLValueType.STRING, fillUnitPath, "#fillType", "Moisture current fillType")
    THUtils.registerXMLPath(schemaSavegame, XMLValueType.FLOAT, fillUnitPath, "#moisture", "Current moisture percentage")
end
function THMoistureFillUnit.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THVehicle, specializations)
        and SpecializationUtil.hasSpecialization(FillUnit, specializations)
end
function THMoistureFillUnit.registerEvents(vehicleType)
end
function THMoistureFillUnit.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onReadStream", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onChangedFillType", THMoistureFillUnit)
    SpecializationUtil.registerEventListener(vehicleType, "onFillUnitFillLevelChanged", THMoistureFillUnit)
end
function THMoistureFillUnit.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "thAA_readFillUnitMoistureFromStream", THMoistureFillUnit.thAA_readFillUnitMoistureFromStream)
    SpecializationUtil.registerFunction(vehicleType, "thAA_writeFillUnitMoistureToStream", THMoistureFillUnit.thAA_writeFillUnitMoistureToStream)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getFillUnitMoistureData", THMoistureFillUnit.thAA_getFillUnitMoistureData)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getFillUnitMoisture", THMoistureFillUnit.thAA_getFillUnitMoisture)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getFillTypeMoisture", THMoistureFillUnit.thAA_getFillTypeMoisture)
    SpecializationUtil.registerFunction(vehicleType, "thAA_setFillUnitMoisture", THMoistureFillUnit.thAA_setFillUnitMoisture)
    SpecializationUtil.registerFunction(vehicleType, "thAA_updateFillUnitMoisture", THMoistureFillUnit.thAA_updateFillUnitMoisture)
    SpecializationUtil.registerFunction(vehicleType, "thAA_setFillUnitInputMoisture", THMoistureFillUnit.thAA_setFillUnitInputMoisture)
    SpecializationUtil.registerFunction(vehicleType, "thAA_getFillUnitInputMoisture", THMoistureFillUnit.thAA_getFillUnitInputMoisture)
end
function THMoistureFillUnit.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "addFillUnitFillLevel", THMoistureFillUnit.addFillUnitFillLevel)
end
function THMoistureFillUnit:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.moistureFillUnits = {}
    specTable.inputMoisture = {}
end
function THMoistureFillUnit:onLoad(savegame)
    local specTable = getSpecTable(self)
    local fillUnitsArray = self:getFillUnits()
    local numFillUnits = #fillUnitsArray
    if numFillUnits > 0 then
        for fillUnitIndex = 1, numFillUnits do
            local fillUnitMoistureData = {
                index = fillUnitIndex,
                currentMoisture = -1,
                lastSentMoisture = -1,
                moistureFillType = FillType.UNKNOWN
            }
            specTable.moistureFillUnits[fillUnitIndex] = fillUnitMoistureData
        end
    end
    specTable.moistureDirtyFlag = self:getNextDirtyFlag()
end
function THMoistureFillUnit:onPostLoad(savegame)
    if self.isServer and savegame ~= nil then
        local xmlFile = savegame.xmlFile
        local xmlKey = THVehicle.getSavegameBaseKey(THMoistureFillUnit.SPEC_NAME, savegame)
        xmlFile:iterate(xmlKey .. ".fillUnit", function(_, fillUnitKey)
            local fillUnitIndex = THUtils.getXMLValue(xmlFile, fillUnitKey, "#index")
            local fillTypeName = THUtils.getXMLValue(xmlFile, fillUnitKey, "#fillType")
            local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
            if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
                local fillTypeIndex = fillTypeInfo.index
                local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex)
                local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex)
                if fillUnitFillLevel ~= nil and fillUnitFillLevel > 0 and fillUnitFillType == fillTypeIndex then
                    local fillUnitMoisture = THUtils.getXMLValue(xmlFile, fillUnitKey, "#moisture")
                    if fillUnitMoisture ~= nil then
                        self:thAA_setFillUnitMoisture(fillUnitIndex, fillUnitMoisture, true)
                    end
                end
            end
        end)
    end
end
function THMoistureFillUnit:onUpdate(dt, ...)
    local specTable = getSpecTable(self)
    local hasInputMoisture = false
    for fillUnitIndex, inputMoistureData in pairs(specTable.inputMoisture) do
        if inputMoistureData.resetTime ~= nil and inputMoistureData.resetTime > 0 then
            local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0
            local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex) or FillType.UNKNOWN
            inputMoistureData.resetTime = inputMoistureData.resetTime - dt
            if inputMoistureData.resetTime <= 0
                or fillUnitFillLevel <= 0 or fillUnitFillType ~= inputMoistureData.fillTypeIndex
            then
                THUtils.clearTable(inputMoistureData)
            else
                hasInputMoisture = true
            end
        end
    end
    if hasInputMoisture then
        self:raiseActive()
    end
end
function THMoistureFillUnit:saveToXMLFile(xmlFile, xmlKey, ...)
    if self.isServer then
        THUtils.call(function()
            local fillUnitsArray = self:getFillUnits()
            local numFillUnits = #fillUnitsArray
            local fillUnitXMLIndex = 0
            if numFillUnits > 0 then
                for fillUnitIndex = 1, numFillUnits do
                    local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
                    local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex)
                    local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex)
                    if fillUnitMoistureData ~= nil
                        and fillUnitFillLevel ~= nil and fillUnitFillLevel > 0
                        and fillUnitFillType ~= nil and fillUnitFillType ~= FillType.UNKNOWN
                    then
                        local fillTypeInfo = thFillTypeManager:getFillType(fillUnitFillType)
                        local fillUnitKey = string.format("%s.fillUnit(%d)", xmlKey, fillUnitXMLIndex)
                        local isBaseEntryAdded = false
                        if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
                            local fillTypeIndex = fillTypeInfo.index
                            local function addFillUnitBaseEntry()
                                if not isBaseEntryAdded then
                                    xmlFile:setInt(fillUnitKey .. "#index", fillUnitIndex)
                                    xmlFile:setString(fillUnitKey .. "#fillType", fillTypeInfo.name)
                                    isBaseEntryAdded = true
                                end
                            end
                            local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillTypeIndex)
                            local fillUnitMoisture = fillUnitMoistureData.currentMoisture
                            if fillTypeMoistureData ~= nil and fillUnitMoisture >= 0 then
                                local minMoisture = fillTypeMoistureData.minPercent
                                local maxMoisture = fillTypeMoistureData.maxPercent
                                addFillUnitBaseEntry()
                                fillUnitMoisture = MathUtil.clamp(fillUnitMoisture, minMoisture, maxMoisture)
                                xmlFile:setFloat(fillUnitKey .. "#moisture", fillUnitMoisture)
                            end
                        end
                        if isBaseEntryAdded then
                            fillUnitXMLIndex = fillUnitXMLIndex + 1
                        end
                    end
                end
            end
        end)
    end
end
function THMoistureFillUnit:onReadStream(streamId, connection)
    if connection:getIsServer() then
        self:thAA_readFillUnitMoistureFromStream(streamId, connection)
    end
end
function THMoistureFillUnit:onWriteStream(streamId, connection)
    if not connection:getIsServer() then
        self:thAA_writeFillUnitMoistureToStream(streamId, connection)
    end
end
function THMoistureFillUnit:onReadUpdateStream(streamId, timeStamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isEnabled/isDirty
            self:thAA_readFillUnitMoistureFromStream(streamId, connection)
        end
    end
end
function THMoistureFillUnit:onWriteUpdateStream(streamId, connection, dirtyMask)
    local specTable = getSpecTable(self)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, specTable.moistureDirtyFlag) ~= 0) then
            self:thAA_writeFillUnitMoistureToStream(streamId, connection)
        end
    end
end
function THMoistureFillUnit:onChangedFillType(fillUnitIndex, fillTypeIndex, oldFillTypeIndex)
    self:thAA_updateFillUnitMoisture(fillUnitIndex)
end
function THMoistureFillUnit:onFillUnitFillLevelChanged(fillUnitIndex, deltaFill, fillTypeIndex, toolType, fillPosData, appliedDelta)
    self:thAA_updateFillUnitMoisture(fillUnitIndex, fillTypeIndex, nil, appliedDelta)
end
function THMoistureFillUnit:thAA_readFillUnitMoistureFromStream(streamId, connection)
    local fillUnitsArray = self:getFillUnits()
    local numFillUnits = #fillUnitsArray
    if numFillUnits > 0 then
        for fillUnitIndex = 1, numFillUnits do
            local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
            if fillUnitMoistureData ~= nil then
                local fillUnitMoisture = streamReadFloat32(streamId)
                self:thAA_setFillUnitMoisture(fillUnitIndex, fillUnitMoisture, true)
            end
        end
    end
end
function THMoistureFillUnit:thAA_writeFillUnitMoistureToStream(streamId, connection)
    local fillUnitsArray = self:getFillUnits()
    local numFillUnits = #fillUnitsArray
    if numFillUnits > 0 then
        for fillUnitIndex = 1, numFillUnits do
            local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
            if fillUnitMoistureData ~= nil then
                local fillUnitMoisture = THUtils.getNoNil(fillUnitMoistureData.currentMoisture, -1)
                streamWriteFloat32(streamId, fillUnitMoisture)
                fillUnitMoistureData.lastSentMoisture = fillUnitMoisture
            end
        end
    end
end
function THMoistureFillUnit:thAA_getFillUnitMoistureData(fillUnitIndex)
    local specTable = getSpecTable(self)
    if fillUnitIndex ~= nil then
        return specTable.moistureFillUnits[fillUnitIndex]
    end
end
function THMoistureFillUnit:thAA_getFillUnitMoisture(fillUnitIndex)
    local fillUnitMoisture = -1
    local fillUnitFillLevel = 0
    local fillUnitFillType = FillType.UNKNOWN
    local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
    if fillUnitMoistureData ~= nil then
        fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0
        fillUnitFillType = self:getFillUnitFillType(fillUnitIndex) or FillType.UNKNOWN
        if fillUnitFillLevel > 0 and fillUnitFillType ~= FillType.UNKNOWN then
            local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillUnitFillType)
            if fillTypeMoistureData ~= nil then
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local defMoisture = fillTypeMoistureData.defPercent
                local curMoisture = THUtils.getNoNil(fillUnitMoistureData.currentMoisture, -1)
                if curMoisture < 0 then
                    fillUnitMoisture = defMoisture
                else
                    fillUnitMoisture = MathUtil.clamp(curMoisture, minMoisture, maxMoisture)
                end
            end
        end
    end
    return fillUnitMoisture, fillUnitFillLevel, fillUnitFillType
end
function THMoistureFillUnit:thAA_getFillTypeMoisture(fillType)
    local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillType)
    local totalMoisture = -1
    local totalFillLevel = 0
    if fillTypeMoistureData ~= nil then
        local fillTypeIndex = fillTypeMoistureData.fillType.index
        local minMoisture = fillTypeMoistureData.minPercent
        local maxMoisture = fillTypeMoistureData.maxPercent
        local fillUnitsArray = self:getFillUnits()
        local numFillUnits = #fillUnitsArray
        if numFillUnits > 0 then
            for fillUnitIndex = 1, numFillUnits do
                local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
                if fillUnitMoistureData ~= nil then
                    local fillUnitMoisture, fillUnitFillLevel, fillUnitFillType = self:thAA_getFillUnitMoisture(fillUnitIndex)
                    if fillUnitFillLevel > 0 and fillUnitFillType == fillTypeIndex then
                        totalFillLevel = totalFillLevel + fillUnitFillLevel
                        if totalMoisture < 0 then
                            totalMoisture = fillUnitMoisture
                        else
                            local newFillPercent = MathUtil.clamp(fillUnitFillLevel / totalFillLevel, 0, 1)
                            local oldFillPercent = 1 - newFillPercent
                            totalMoisture = MathUtil.clamp((totalMoisture * oldFillPercent) + (fillUnitMoisture * newFillPercent), minMoisture, maxMoisture)
                        end
                    end
                end
            end
        end
    end
    return totalMoisture, totalFillLevel
end
function THMoistureFillUnit:thAA_setFillUnitMoisture(fillUnitIndex, moisture, noEventSend)
    local specTable = getSpecTable(self)
    THUtils.call(function()
        if THUtils.argIsValid(type(moisture) == "number", "moisture", moisture) then
            local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
            if fillUnitMoistureData ~= nil then
                local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0
                local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex) or FillType.UNKNOWN
                local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillUnitFillType)
                local oldMoisture = fillUnitMoistureData.currentMoisture
                local newMoisture = oldMoisture
                fillUnitMoistureData.moistureFillType = fillUnitFillType
                if fillTypeMoistureData == nil or fillUnitFillLevel <= 0 or fillUnitFillType == FillType.UNKNOWN then
                    newMoisture = -1
                else
                    local minMoisture = fillTypeMoistureData.minPercent
                    local maxMoisture = fillTypeMoistureData.maxPercent
                    local defMoisture = fillTypeMoistureData.defPercent
                    if moisture < 0 then
                        newMoisture = defMoisture
                    else
                        newMoisture = MathUtil.clamp(moisture, minMoisture, maxMoisture)
                    end
                end
                if math.abs(newMoisture - oldMoisture) > 0.000001 then
                    if THDebugUtil.getIsEnabled(debugFlagId) then
                        THDebugUtil.printMsg(debugFlagId, nil, "Updating fillUnit: %s", fillUnitIndex)
                        THDebugUtil.printMsg(debugFlagId, nil, "- Old moisture: %s", oldMoisture)
                        THDebugUtil.printMsg(debugFlagId, nil, "- New moisture: %s", newMoisture)
                        THDebugUtil.printMsg(debugFlagId, nil, "")
                    end
                    fillUnitMoistureData.currentMoisture = newMoisture
                end
                if self.isServer and noEventSend ~= true then
                    local currentMoisture = fillUnitMoistureData.currentMoisture or -1
                    local lastSentMoisture = fillUnitMoistureData.lastSentMoisture
                    if lastSentMoisture == nil or math.abs(currentMoisture - lastSentMoisture) > 0.0001 then
                        self:raiseDirtyFlags(specTable.moistureDirtyFlag)
                    end
                end
                return true
            end
        end
    end)
    return false
end
function THMoistureFillUnit:thAA_updateFillUnitMoisture(fillUnitIndex, inputFillType, inputMoisture, inputDelta, noEventSend)
    local specTable = getSpecTable(self)
    if self.isServer then
        local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
        if fillUnitMoistureData == nil and fillUnitIndex ~= nil then
            specTable.inputMoisture[fillUnitIndex] = nil
        elseif fillUnitMoistureData ~= nil then
            local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0
            local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex) or FillType.UNKNOWN
            local moistureFillType = fillUnitMoistureData.moistureFillType or FillType.UNKNOWN
            local oldMoisture = fillUnitMoistureData.currentMoisture
            local newMoisture = oldMoisture
            local fillUnitFillTypeName = thFillTypeManager:getFillTypeName(fillUnitFillType)
            local moistureFillTypeName = thFillTypeManager:getFillTypeName(moistureFillType)
            local inputFillTypeName = thFillTypeManager:getFillTypeName(inputFillType)
            fillUnitMoistureData.moistureFillType = fillUnitFillType
            if fillUnitFillLevel <= 0 or fillUnitFillType == FillType.UNKNOWN then
                newMoisture = -1
            else
                local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillUnitFillType)
                if fillTypeMoistureData == nil then
                    newMoisture = -1
                else
                    local minMoisture = fillTypeMoistureData.minPercent
                    local maxMoisture = fillTypeMoistureData.maxPercent
                    local defMoisture = fillTypeMoistureData.defPercent
                    if fillUnitFillType ~= moistureFillType then
                        oldMoisture = -1 -- reset
                    end
                    inputMoisture = inputMoisture or self:thAA_getFillUnitInputMoisture(fillUnitIndex)
                    inputDelta = inputDelta or 0
                    if inputMoisture ~= nil and inputMoisture >= 0 and inputDelta > 0
                        and inputFillType ~= nil and inputFillType ~= FillType.UNKNOWN
                    then
                        if inputFillType ~= fillUnitFillType then
                            if THDebugUtil.getIsEnabled(debugFlagId) then
                                THDebugUtil.printMsg(debugFlagId, nil, "Input fillType (%s) does not match fillUnit %d fillType (%s)", inputFillTypeName, fillUnitIndex, fillUnitFillTypeName)
                            end
                        else
                            inputMoisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                            if oldMoisture < 0 then
                                newMoisture = inputMoisture
                            else
                                local newFillPercent = MathUtil.clamp(inputDelta / fillUnitFillLevel, 0, 1)
                                local oldFillPercent = 1 - newFillPercent
                                newMoisture = (oldMoisture * oldFillPercent) + (inputMoisture * newFillPercent)
                            end
                        end
                    elseif oldMoisture < 0 then
                        newMoisture = defMoisture
                    elseif inputDelta > 0 then
                        local newFillPercent = MathUtil.clamp(inputDelta / fillUnitFillLevel, 0, 1)
                        local oldFillPercent = 1 - newFillPercent
                        newMoisture = (oldMoisture * oldFillPercent) + (defMoisture * newFillPercent)
                    end
                    if newMoisture ~= oldMoisture then
                        newMoisture = MathUtil.clamp(newMoisture, minMoisture, maxMoisture)
                    end
                end
            end
            if math.abs(newMoisture - oldMoisture) > 0.000001 then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THDebugUtil.printMsg(debugFlagId, nil, "Updating fillUnit: %d", fillUnitIndex)
                    if inputFillType ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input fillType: %s", inputFillTypeName)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "- Old moisture: %s", oldMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "- New Moisture: %s", newMoisture)
                    if inputMoisture ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input moisture: %s", inputMoisture)
                    end
                    if inputDelta ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input delta: %s", inputDelta)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                    if fillUnitFillType ~= moistureFillType then
                        THDebugUtil.printMsg(debugFlagId, nil, "FillType changed from %q to %q", moistureFillTypeName, fillUnitFillTypeName)
                        THDebugUtil.printMsg(debugFlagId, nil, "")
                    end
                end
                fillUnitMoistureData.currentMoisture = newMoisture
            end
            if noEventSend ~= true then
                local currentMoisture = fillUnitMoistureData.currentMoisture or -1
                local lastSentMoisture = fillUnitMoistureData.lastSentMoisture
                if lastSentMoisture == nil or math.abs(currentMoisture - lastSentMoisture) > 0.0001 then
                    self:raiseDirtyFlags(specTable.moistureDirtyFlag)
                end
            end
        end
        return true
    end
    return false
end
function THMoistureFillUnit:thAA_getFillUnitInputMoisture(fillUnitIndex)
    local specTable = getSpecTable(self)
    local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
    if fillUnitMoistureData ~= nil then
        local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0
        local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex) or FillType.UNKNOWN
        local inputMoistureData = specTable.inputMoisture[fillUnitIndex]
        if inputMoistureData ~= nil then
            if fillUnitFillLevel <= 0 or fillUnitFillType == FillType.UNKNOWN
                or fillUnitFillType == inputMoistureData.fillTypeIndex
            then
                return inputMoistureData.moisture
            end
        end
    end
end
function THMoistureFillUnit:thAA_setFillUnitInputMoisture(fillUnitIndex, fillType, inputMoisture)
    local specTable = getSpecTable(self)
    if THUtils.argIsValid(fillUnitIndex == nil or type(fillUnitIndex) == "number", "fillUnitIndex", fillUnitIndex)
        and THUtils.argIsValid(type(inputMoisture) == "number", "inputMoisture", inputMoisture)
    then
        local fillUnitMoistureData = self:thAA_getFillUnitMoistureData(fillUnitIndex)
        local fillTypeMoistureData = g_thMoistureSystem:getFillTypeMoistureData(fillType)
        if fillUnitMoistureData == nil or fillTypeMoistureData == nil then
            if fillUnitIndex ~= nil then
                specTable.inputMoisture[fillUnitIndex] = nil
            end
        else
            local inputMoistureData = THUtils.createSubTable(specTable.inputMoisture, fillUnitIndex, true)
            local minMoisture = fillTypeMoistureData.minPercent
            local maxMoisture = fillTypeMoistureData.maxPercent
            if inputMoisture ~= nil and inputMoisture >= 0 then
                local fillUnitFillLevel = self:getFillUnitFillLevel(fillUnitIndex) or 0
                local fillUnitFillType = self:getFillUnitFillType(fillUnitIndex) or FillType.UNKNOWN
                local inputFillTypeIndex = fillTypeMoistureData.fillType.index
                if fillUnitFillLevel <= 0 or fillUnitFillType == FillType.UNKNOWN
                    or fillUnitFillType == inputFillTypeIndex
                then
                    inputMoisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                    inputMoistureData.fillUnitIndex = fillUnitIndex
                    inputMoistureData.fillTypeIndex = inputFillTypeIndex
                    inputMoistureData.moisture = inputMoisture
                    inputMoistureData.resetTime = 50
                    self:raiseActive()
                end
            end
        end
    end
end
function THMoistureFillUnit:addFillUnitFillLevel(superFunc, farmId, fillUnitIndex, deltaFill, fillTypeIndex, ...)
    THUtils.call(function()
        local inputMoisture = self:thAA_getFillUnitInputMoisture(fillUnitIndex)
        if inputMoisture ~= nil and inputMoisture >= 0
            and deltaFill ~= nil and deltaFill > 0
        then
            local _, yieldFactor = g_thMoistureSystem:getFillTypeMoistureFactors(fillTypeIndex, inputMoisture)
            deltaFill = math.max(0.001, deltaFill * yieldFactor)
        end
    end)
    return superFunc(self, farmId, fillUnitIndex, deltaFill, fillTypeIndex, ...)
end